"""
Some `xarray accessors
<https://docs.xarray.dev/en/stable/internals/extending-xarray.html>`_ for convenient
data analysis and visualization.

.. currentmodule:: erlab.accessors

"""

__all__ = [
    "PlotAccessor",
    "ImageToolAccessor",
    "SelectionAccessor",
    "MomentumAccessor",
    "OffsetView",
]

import functools
import time
import warnings
from collections.abc import Callable, ItemsView, Iterable, Iterator
from typing import Literal

import numpy as np
import numpy.typing as npt
import xarray as xr
from xarray.core.utils import either_dict_or_kwargs

import erlab.plotting.erplot as eplt
from erlab.analysis.interpolate import interpn
from erlab.analysis.kspace import AxesConfiguration, get_kconv_func, kz_func
from erlab.constants import rel_kconv, rel_kzconv
from erlab.interactive.imagetool import ImageTool, itool
from erlab.interactive.kspace import ktool


class ERLabAccessor:
    """Base class for accessors."""

    def __init__(self, xarray_obj: xr.DataArray | xr.Dataset):
        self._obj = xarray_obj


@xr.register_dataarray_accessor("qplot")
class PlotAccessor(ERLabAccessor):
    """`xarray.DataArray.qplot` accessor for plotting data."""

    def __call__(self, *args, **kwargs):
        """
        Plot the data. If a 2D data array is provided, it is plotted using
        :func:`plot_array <erlab.plotting.general.plot_array>`. Otherwise, it is
        equivalent to calling :meth:`xarray.DataArray.plot`.

        Parameters
        ----------
        *args
            Positional arguments to be passed to the plotting function.
        **kwargs
            Keyword arguments to be passed to the plotting function.

        """
        if len(self._obj.dims) == 2:
            return eplt.plot_array(self._obj, *args, **kwargs)
        else:
            return self._obj.plot(*args, **kwargs)


@xr.register_dataarray_accessor("qshow")
class ImageToolAccessor(ERLabAccessor):
    """`xarray.DataArray.qshow` accessor for interactive visualization."""

    def __call__(self, *args, **kwargs) -> ImageTool:
        if len(self._obj.dims) >= 2:
            return itool(self._obj, *args, **kwargs)
        else:
            raise ValueError("Data must have at leasst two dimensions.")


@xr.register_dataarray_accessor("qsel")
class SelectionAccessor(ERLabAccessor):
    """
    `xarray.DataArray.qsel` accessor for conveniently selecting and averaging
    data.
    """

    def __call__(
        self,
        indexers: dict[str, float | slice] | None = None,
        *,
        verbose: bool = False,
        **indexers_kwargs: dict[str, float | slice],
    ):
        """Select and average data along specified dimensions.

        Parameters
        ----------
        indexers
            Dictionary specifying the dimensions and their values or slices.
            Position along a dimension can be specified in three ways:

            - As a scalar value: `alpha=-1.2`

              If no width is specified, the data is selected along the nearest value. It
              is equivalent to `xarray.DataArray.sel` with `method='nearest'`.

            - As a value and width: `alpha=5, alpha_width=0.5`

              The data is *averaged* over a slice of width `alpha_width`, centered at
              `alpha`.

            - As a slice: `alpha=slice(-10, 10)`

              The data is selected over the specified slice. No averaging is performed.

            One of `indexers` or `indexers_kwargs` must be provided.
        verbose
            If `True`, print information about the selected data and averaging process.
            Default is `False`.
        **indexers_kwargs
            The keyword arguments form of `indexers`. One of `indexers` or
            `indexers_kwargs` must be provided.

        Returns
        -------
        xarray.DataArray
            The selected and averaged data.

        Raises
        ------
        ValueError
            If a specified dimension is not present in the data.
        """

        indexers = either_dict_or_kwargs(indexers, indexers_kwargs, "qsel")

        # Bin widths for each dimension, zero if width not specified
        bin_widths: dict[str, float] = {}

        for dim in indexers:
            if not dim.endswith("_width"):
                bin_widths[dim] = indexers.get(f"{dim}_width", 0.0)
                if dim not in self._obj.dims:
                    raise ValueError(f"Dimension `{dim}` not found in data.")

        scalars: dict[str, float] = {}
        slices: dict[str, slice] = {}
        avg_dims: list[str] = []

        for dim, width in bin_widths.items():
            if width == 0.0:
                if isinstance(indexers[dim], slice):
                    slices[dim] = indexers[dim]
                else:
                    scalars[dim] = float(indexers[dim])
            else:
                slices[dim] = slice(
                    indexers[dim] - width / 2, indexers[dim] + width / 2
                )
                avg_dims.append(dim)

        if len(scalars) >= 1:
            out = self._obj.sel(**scalars, method="nearest")
        else:
            out = self._obj

        if len(slices) >= 1:
            out = out.sel(**slices)

            lost_coords = {k: out[k].mean() for k in avg_dims}
            out = out.mean(dim=avg_dims, keep_attrs=True)
            out = out.assign_coords(lost_coords)

        if verbose:
            print(
                f"Selected data with {scalars} and {slices}, averaging over {avg_dims}"
            )

        return out


def only_angles(method: Callable | None = None):
    """
    A decorator that ensures the data is in angle space before executing the decorated
    method.

    If the data is not in angle space (i.e., if "kx" or "ky" dimensions are present), a
    `ValueError` is raised.
    """

    def wrapper(method: Callable):
        @functools.wraps(method)
        def _impl(self, *args, **kwargs):
            if "kx" in self._obj.dims or "ky" in self._obj.dims:
                raise ValueError(
                    f"`{method.__name__}` cannot be called for data in momentum space."
                )
            return method(self, *args, **kwargs)

        return _impl

    if method is not None:
        return wrapper(method)
    return wrapper


def only_momentum(method: Callable | None = None):
    """
    A decorator that ensures the data is in momentum space before executing the
    decorated method.

    If the data is not in momentum space (i.e., if "kx" nor "ky" dimensions are
    present), a `ValueError` is raised.
    """

    def wrapper(method: Callable):
        @functools.wraps(method)
        def _impl(self, *args, **kwargs):
            if not ("kx" in self._obj.dims or "ky" in self._obj.dims):
                raise ValueError(
                    f"`{method.__name__}` cannot be called for data in angle space."
                )
            return method(self, *args, **kwargs)

        return _impl

    if method is not None:
        return wrapper(method)
    return wrapper


class OffsetView:
    r"""A class representing an offset view for an `xarray.DataArray`.

    This class provides a convenient way to access and manipulate angle offsets
    associated with the given data.

    Parameters
    ----------
    xarray_obj
        The `xarray.DataArray` for which the offset view is created.

    Methods
    -------
    __len__() -> int:
        Returns the number of valid offset keys.

    __iter__() -> Iterator[str, float]:
        Returns an iterator over the valid offset keys and their corresponding values.

    __getitem__(key: str) -> float:
        Returns the offset value associated with the given key.

    __setitem__(key: str, value: float) -> None:
        Sets the offset value for the given key.

    __eq__(other: object) -> bool:
        Compares the offset view with another object for equality. `True` if the
        dictionary representation is equal, `False` otherwise.

    __repr__() -> str:
        Returns a string representation of the offset view.

    _repr_html_() -> str:
        Returns an HTML representation of the offset view.
    """

    def __init__(self, xarray_obj: xr.DataArray):
        self._obj = xarray_obj
        for k in self._obj.kspace.valid_offset_keys:
            if k + "_offset" not in self._obj.attrs:
                self[k] = 0.0

    def __len__(self) -> int:
        return len(self._obj.kspace.valid_offset_keys)

    def __iter__(self) -> Iterator[str, float]:
        for key in self._obj.kspace.valid_offset_keys:
            yield key, self.__getitem__(key)

    def __getitem__(self, key: str) -> float:
        if key in self._obj.kspace.valid_offset_keys:
            return float(self._obj.attrs[key + "_offset"])
        else:
            raise KeyError(
                f"Invalid offset key `{key}` for experimental configuration "
                f"{self._obj.kspace.configuration}"
            )

    def __setitem__(self, key: str, value: float) -> None:
        if key in self._obj.kspace.valid_offset_keys:
            self._obj.attrs[key + "_offset"] = float(value)

    def __eq__(self, other: object) -> bool:
        return dict(self) == dict(other)

    def __repr__(self) -> str:
        return dict(self).__repr__()

    def _repr_html_(self) -> str:
        out = ""
        out += "<table><tbody>"
        for k, v in self.items():
            out += (
                "<tr>"
                f"<td style='text-align:left;'><b>{k}</b></td>"
                f"<td style='text-align:left;'>{v}</td>"
                "</tr>"
            )
        out += "</tbody></table>"
        return out

    def update(
        self,
        other: dict | Iterable[tuple[str, float]] | None = None,
        **kwargs: dict[str, float],
    ) -> "OffsetView":
        """Updates the offset view with the provided key-value pairs."""
        if other is not None:
            for k, v in other.items() if isinstance(other, dict) else other:
                self[k] = v
        for k, v in kwargs.items():
            self[k] = v
        return self

    def items(self) -> ItemsView[str, float]:
        """Returns a view of the offset view as a list of (key, value) pairs."""
        return dict(self).items()

    def reset(self) -> "OffsetView":
        """Reset all angle offsets to zero."""
        for k in self._obj.kspace.valid_offset_keys:
            self[k] = 0.0
        return self


@xr.register_dataarray_accessor("kspace")
class MomentumAccessor:
    """`xarray.DataArray.kspace` accessor for momentum conversion related utilities.

    This class provides convenient access to various momentum-related properties of a
    data object. It allows getting and setting properties such as configuration, inner
    potential, work function, angle resolution, slit axis, momentum axes, angle
    parameters, and offsets.

    """

    def __init__(self, xarray_obj: xr.DataArray):
        self._obj = xarray_obj

    @property
    def configuration(self) -> AxesConfiguration:
        """Returns the experimental configuration.

        For a properly implemented data loader, the configuration attribute must be set
        on data import. See :class:`erlab.analysis.kspace.AxesConfiguration` for
        details.
        """
        if "configuration" not in self._obj.attrs:
            raise ValueError(
                "Configuration not found in data attributes! "
                "Data attributes may have been discarded since initial import."
            )
        return AxesConfiguration(int(self._obj.attrs.get("configuration")))

    @configuration.setter
    def configuration(self, value: AxesConfiguration | int):
        self._obj.attrs["configuration"] = int(value)

    @property
    def inner_potential(self) -> float:
        """Inner potential of the sample in eV.

        The inner potential is stored in the ``inner_potential`` attribute of the data.
        If the inner potential is not set, a warning is issued and a default value of
        10.0 eV is assumed.

        Note
        ----
        This property provides a setter method that takes a float value and sets the
        data attribute accordingly.

        Example
        -------
        >>> data.kspace.inner_potential = 13.0
        >>> data.kspace.inner_potential
        13.0
        """

        if "inner_potential" in self._obj.attrs:
            return float(self._obj.attrs["inner_potential"])
        else:
            warnings.warn(
                "Inner potential not found in data attributes, assuming 10 eV",
                stacklevel=1,
            )
            return 10.0

    @inner_potential.setter
    def inner_potential(self, value: float):
        self._obj.attrs["inner_potential"] = float(value)

    @property
    def work_function(self) -> float:
        """Work function of the sample in eV.

        The work function is stored in the ``sample_workfunction`` attribute of the
        data. If the work function is not set, a warning is issued and a default value
        of 4.5 eV is assumed.

        Note
        ----
        This property provides a setter method that takes a float value and sets the
        data attribute accordingly.

        Example
        -------
        >>> data.kspace.work_function = 4.5
        >>> data.kspace.work_function
        4.5
        """

        if "sample_workfunction" in self._obj.attrs:
            return float(self._obj.attrs["sample_workfunction"])
        else:
            warnings.warn(
                "Work function not found in data attributes, assuming 4.5 eV",
                stacklevel=1,
            )
            return 4.5

    @work_function.setter
    def work_function(self, value: float):
        self._obj.attrs["sample_workfunction"] = float(value)

    @property
    def angle_resolution(self) -> float:
        """Angular resolution of the data in degrees.

        The angular resolution is stored in the ``angle_resolution`` attribute of the
        data. If it is not set, a default value of 0.1° is silently used. It is used in
        `best_kp_resolution` when automatically estimating momentum steps through
        `estimate_resolution`.
        """

        try:
            return float(self._obj.attrs["angle_resolution"])
        except KeyError:
            # warnings.warn(
            #     "Angle resolution not found in data attributes, assuming 0.1 degrees"
            # )
            return 0.1

    @angle_resolution.setter
    def angle_resolution(self, value: float):
        self._obj.attrs["angle_resolution"] = float(value)

    @property
    def slit_axis(self) -> str:
        """Returns the momentum axis parallel to the slit.

        Returns
        -------
        str
            Returns ``'kx'`` for type 1 configurations, ``'ky'`` otherwise.
        """
        match self.configuration:
            case AxesConfiguration.Type1 | AxesConfiguration.Type1DA:
                return "kx"
            case _:
                return "ky"

    @property
    @only_angles
    def momentum_axes(self) -> tuple[str, ...]:
        """Returns the momentum axes of the data after conversion.

        Returns
        -------
        tuple
            For photon energy dependent scans, it returns the slit axis and ``'kz'``.
            For maps, it returns ``'kx'`` and ``'ky'``. Otherwise, it returns only the
            slit axis.

        """
        if self.has_hv:
            return (self.slit_axis, "kz")
        elif self.has_beta:
            return ("kx", "ky")
        else:
            return (self.slit_axis,)

    @property
    def angle_params(self) -> dict[str, float]:
        """Parameters passed to :func:`erlab.analysis.kspace.get_kconv_func`."""
        params = {
            "delta": self.offsets["delta"],
            "xi": float(self._obj["xi"].values),
            "xi0": self.offsets["xi"],
        }
        match self.configuration:
            case AxesConfiguration.Type1 | AxesConfiguration.Type2:
                params["beta0"] = self.offsets["beta"]
            case _:
                params["chi"] = float(self._obj["chi"].values)
                params["chi0"] = self.offsets["chi"]
        return params

    @property
    def _photon_energy(self) -> npt.NDArray[np.floating] | float:
        # Make photon energy axis always comes last
        if self.has_hv:
            return self._obj.hv.values[
                tuple(
                    slice(None) if (i == self._obj.ndim - 1) else np.newaxis
                    for i in range(self._obj.ndim)
                )
            ]
        else:
            # Scalar
            return float(self._obj.hv)

    @property
    def _beta(self) -> npt.NDArray[np.floating] | float:
        if self.has_beta:
            if self.has_hv:
                # Assume hv-dependent beta, make same shape as hv (last axis)
                return self._obj["beta"].values[
                    tuple(
                        slice(None) if (i == self._obj.ndim - 1) else np.newaxis
                        for i in range(self._obj.ndim)
                    )
                ]
            else:
                # First axis
                return self._obj["beta"].values[
                    tuple(
                        slice(None) if (i == 0) else np.newaxis
                        for i in range(self._obj.ndim)
                    )
                ]
        else:
            # Scalar
            return float(self._obj["beta"])

    @property
    @only_angles
    def _alpha(self) -> npt.NDArray[np.floating] | float:
        if not self.has_beta and not self.has_hv:
            return self._obj["alpha"].values
        else:
            return self._obj["alpha"].values[
                tuple(
                    slice(None) if (i == 0) else np.newaxis
                    for i in range(self._obj.ndim)
                )
            ]

    @property
    def _binding_energy(self) -> npt.NDArray[np.floating] | float:
        if self.has_eV:
            return self._obj.eV.values[:, np.newaxis]
        else:
            return float(self._obj.eV.values)

    @property
    def _kinetic_energy(self) -> npt.NDArray[np.floating] | float:
        return self._photon_energy - self.work_function + self._binding_energy

    @property
    def has_eV(self) -> bool:
        """Returns `True` if object has an energy axis."""
        return "eV" in self._obj.dims

    @property
    @only_angles
    def has_hv(self) -> bool:
        """Returns `True` for photon energy dependent data."""
        return self._obj["hv"].size > 1

    @property
    @only_angles
    def has_beta(self) -> bool:
        """Check if the coordinate array for :math:`β` has more than one element.

        Returns
        -------
        bool
            Returns `True` if the size of the coordinate array for :math:`β` is greater
            than 1, `False` otherwise.

        Note
        ----
        This may be `True` for data that are not maps. For instance,
        :math:`hν`-dependent cuts with an in-plane momentum offset may have a
        :math:`hν`-dependent :math:`β` offset.

        """
        return self._obj.beta.size > 1

    @property
    def valid_offset_keys(self) -> tuple[str, ...]:
        """
        Get valid offset angles based on the experimental configuration.

        Returns
        -------
        tuple
            A tuple containing the valid offset keys. For configurations with a
            deflector, returns ``("delta", "chi", "xi")``. Otherwise, returns
            ``("delta", "xi", "beta")``.
        """
        match self.configuration:
            case AxesConfiguration.Type1 | AxesConfiguration.Type2:
                return ("delta", "xi", "beta")
            case _:
                return ("delta", "chi", "xi")

    @property
    def offsets(self) -> OffsetView:
        """Angle offsets used in momentum conversion.

        Returns
        -------
        OffsetView
            A mapping between valid offset keys and their corresponding offsets.

        Examples
        --------

        - View all offsets

          >>> data.kspace.offsets
          {'delta': 0.0, 'xi': 0.0, 'beta': 0.0}

        - View single offset

          >>> data.kspace.offsets["beta"]
          0.0

        - Offsets to dictionary

          >>> dict(data.kspace.offsets)
          {'delta': 0.0, 'xi': 0.0, 'beta': 0.0}

        - Set single offset

          >>> data.kspace.offsets["beta"] = 3.0
          >>> data.kspace.offsets
          {'delta': 0.0, 'xi': 0.0, 'beta': 3.0}

        - Overwrite offsets with dictionary

          >>> data.kspace.offsets = dict(delta=1.5, xi=2.7)
          >>> data.kspace.offsets
          {'delta': 1.5, 'xi': 2.7, 'beta': 0.0}

        - Update offsets

          >>> data.kspace.offsets.update(beta=0.1, xi=0.0)
          {'delta': 1.5, 'xi': 0.0, 'beta': 0.1}

        - Reset all offsets

          >>> data.kspace.offsets.reset()
          {'delta': 0.0, 'xi': 0.0, 'beta': 0.0}
        """
        if not hasattr(self, "_offsetview"):
            self._offsetview = OffsetView(self._obj)

        return self._offsetview

    @offsets.setter
    def offsets(self, offset_dict: dict[str, float]):
        if not hasattr(self, "_offsetview"):
            self._offsetview = OffsetView(self._obj)

        self._offsetview.reset()
        self._offsetview.update(offset_dict)

    def estimate_bounds(self) -> dict[str, tuple[float, float]]:
        """
        Estimates the bounds of the data in momentum space based on the available
        parameters.

        Returns
        -------
        bounds : dict[str, tuple[float, float]]
            A dictionary containing the estimated bounds for each parameter. The keys of
            the dictionary are 'kx', 'ky', and 'kz' (for :math:`hν`-dependent data). The
            values are tuples representing the minimum and maximum values.

        """
        if self.has_hv:
            # hv dependent cut
            kx, ky = self._forward_func_raw(self._alpha, self._beta)
            kz = kz_func(self._kinetic_energy, self.inner_potential, kx, ky)

            return {
                "kx": (kx.min(), kx.max()),
                "ky": (ky.min(), ky.max()),
                "kz": (kz.min(), kz.max()),
            }

        elif self.has_beta:
            # kxy map
            # construct boundary array from meshgrid
            alpha, beta = (
                np.r_[arr[0, :-1], arr[:-1, -1], arr[-1, ::-1], arr[-2:0:-1, 0]]
                for arr in np.meshgrid(self._obj["alpha"], self._obj["beta"])
            )

            kx, ky = self._forward_func_raw(alpha[np.newaxis, :], beta[np.newaxis, :])

            return {"kx": (kx.min(), kx.max()), "ky": (ky.min(), ky.max())}

        else:
            # just cut
            kx, ky = self._forward_func_raw(self._alpha, self._beta)

            return {"kx": (kx.min(), kx.max()), "ky": (ky.min(), ky.max())}

    @only_angles
    def estimate_resolution(
        self,
        axis: Literal["kx", "ky", "kz"],
        lims: tuple[float, float] | None = None,
        from_numpoints: bool = False,
    ) -> float:
        """Estimate the resolution for a given momentum axis.

        Parameters
        ----------
        axis
            Axis to estimate the resolution for.
        lims
            The limits of the axis used when `from_numpoints` is `True`. If not
            provided, reasonable limits will be calculated by :meth:`estimate_bounds`,
            by default `None`
        from_numpoints
            If `True`, estimate the resolution from the number of points in the relevant
            axis. If `False`, estimate the resolution based on the data, by default
            `False`

        Returns
        -------
        float
            The estimated resolution.

        Raises
        ------
        ValueError
            If no photon energy axis is found in data for axis ``'kz'``.

        """
        if from_numpoints and (lims is None):
            lims = self.estimate_bounds()[axis]

        if axis == self.slit_axis:
            dim = "alpha"

        elif axis == "kz":
            dim = "hv"

            if not self.has_hv:
                raise ValueError("No photon energy axis found.")

        else:
            dim = "beta"

        if from_numpoints:
            return float((lims[1] - lims[0]) / len(self._obj[dim]))
        elif axis == "kz":
            return self.best_kz_resolution
        else:
            return self.best_kp_resolution

    @property
    @only_angles
    def best_kp_resolution(self) -> float:
        r"""
        Estimates the minimum in-plane momentum resolution based on the kinetic energy
        and angular resolution:

        .. math:: \Delta k_{\parallel} \sim \sqrt{2 m_e E_k/\hbar^2} \cos(\alpha) \Delta\alpha

        """
        min_Ek = np.amin(self._kinetic_energy)
        max_angle = max(np.abs(self._obj["alpha"].values))
        return (
            rel_kconv
            * np.sqrt(min_Ek)
            * np.cos(np.deg2rad(max_angle))
            * np.deg2rad(self.angle_resolution)
        )

    @property
    @only_angles
    def best_kz_resolution(self) -> float:
        r"""
        Estimates the minimum out-of-plane momentum resolution based on the mean free
        path :cite:p:`Seah1979` and the kinetic energy.

        .. math:: \Delta k_z \sim 1/\lambda

        """
        kin = self._kinetic_energy.flatten()
        c1, c2 = 641.0, 0.096
        imfp = (c1 / (kin**2) + c2 * np.sqrt(kin)) * 10
        return np.amin(1 / imfp)

    def _forward_func_raw(self, alpha, beta):
        return get_kconv_func(
            self._kinetic_energy, self.configuration, self.angle_params
        )[0](alpha, beta)

    def _inverse_func_raw(self, kx, ky, kz=None):
        return get_kconv_func(
            self._kinetic_energy, self.configuration, self.angle_params
        )[1](kx, ky, kz)

    # def forward(self, phi, theta):
    #     # make alpha and beta (1,N) where N = len(phi) * len(theta)
    #     alpha, beta = np.array(np.meshgrid(phi, theta)).reshape(2, -1)[:, None]

    #     if self.has_hv:
    #         alpha = alpha[:, :, np.newaxis]
    #         beta = beta[:, :, np.newaxis]
    #     # binding energy 없이 photon energy만 주어지는 경우는 없겠지?

    #     return self._forward_func(alpha, beta)

    def _inverse(
        self,
        kx: npt.NDArray[np.floating],
        ky: npt.NDArray[np.floating],
        kz: npt.NDArray[np.floating] | None = None,
    ) -> dict[str, npt.NDArray[np.floating]]:
        if self.has_eV:
            ndim = self._kinetic_energy.ndim
        else:
            ndim = 1
        slicer = tuple(
            slice(None) if (i == ndim - 1) else np.newaxis for i in range(ndim)
        )

        if kz is None:
            # make kxval and kyval (...,N) where N = len(kx) * len(ky)
            kxval, kyval = np.array(np.meshgrid(kx, ky)).reshape(2, -1)
            kxval, kyval = kxval[slicer], kyval[slicer]
            kzval = None
        else:
            kxval, kyval, kzval = np.array(np.meshgrid(kx, ky, kz)).reshape(3, -1)
            kxval, kyval, kzval = kxval[slicer], kyval[slicer], kzval[slicer]

        alpha, beta = self._inverse_func_raw(kxval, kyval, kzval)

        out_dict = {"alpha": alpha, "beta": beta}

        if self.has_eV:
            # Ensure all arrays have the same shape
            binding = self._binding_energy * np.ones_like(alpha)
            for k in out_dict:
                out_dict[k] = out_dict[k] * np.ones_like(binding)
            out_dict["eV"] = binding

        if self.has_hv:
            # Photon energy
            out_dict["hv"] = (
                rel_kzconv * (kxval**2 + kyval**2 + kzval**2)
                - self.inner_potential
                + self.work_function
            )
            if self.has_eV:
                out_dict["hv"] = out_dict["hv"] - out_dict["eV"]
            else:
                out_dict["hv"] = out_dict["hv"] - self._binding_energy

        return out_dict

    @only_angles
    def convert(
        self,
        bounds: dict[str, tuple[float, float]] | None = None,
        resolution: dict[str, float] | None = None,
        *,
        silent: bool = False,
    ) -> xr.DataArray:
        """Convert to momentum space.

        Parameters
        ----------
        bounds
            A dictionary specifying the bounds for each coordinate axis. The keys are
            the names of the axes, and the values are tuples representing the lower and
            upper bounds of the axis. If not provided, the bounds will be estimated
            based on the data.
        resolution
            A dictionary specifying the resolution for each momentum axis. The keys are
            the names of the axes, and the values are floats representing the desired
            resolution of the axis. If not provided, the resolution will be estimated
            based on the data. For in-plane momentum, the resolution is estimated from
            the angle resolution and kinetic energy. For out-of-plane momentum, two
            values are calculated. One is based on the number of photon energy points,
            and the other is estimated as the inverse of the photoelectron inelastic
            mean free path given by the universal curve. The resolution is estimated as
            the smaller of the two values.
        silent
            If `True`, suppresses printing, by default `False`.

        Returns
        -------
        xarray.DataArray
            The converted data.

        Note
        ----
        This method converts the data to a new coordinate system specified by the
        provided bounds and resolution. It uses interpolation to map the data from the
        original coordinate system to the new one.

        The converted data is returned as a DataArray object with updated coordinates
        and dimensions.

        Examples
        --------
        Set parameters and convert with automatic bounds and resolution:

        .. code-block:: python

            data.kspace.offsets = {"delta": 0.1, "xi": 0.0, "beta": 0.3}
            data.kspace.work_function = 4.3
            data.kspace.inner_potential = 12.0
            converted_data = data.kspace.convert()


        Convert with specified bounds and resolution:

        .. code-block:: python

            bounds = {"kx": (0.0, 1.0), "ky": (-1.0, 1.0)}
            resolution = {"kx": 0.01, "ky": 0.01}
            converted_data = data.kspace.convert(bounds, resolution)

        """

        if bounds is None:
            bounds = {}

        if resolution is None:
            resolution = {}

        if not silent:
            print("Estimating bounds and resolution")

        calculated_bounds: dict[str, tuple[float, float]] = self.estimate_bounds()

        new_size: dict[str, int] = {}
        interp_coords: dict[str, np.ndarray] = {}
        other_coords: dict[str, np.ndarray] = {}

        for k, lims in calculated_bounds.items():
            if k in self.momentum_axes:
                lims = bounds.get(k, lims)

                res = self.estimate_resolution(k, lims, from_numpoints=False)
                if k == "kz":
                    res_n = self.estimate_resolution(k, lims, from_numpoints=True)
                    res = min(res, res_n)
                res = resolution.get(k, res)

                new_size[k] = round((lims[1] - lims[0]) / res + 1)

                interp_coords[k] = np.linspace(*lims, new_size[k])
            else:
                # Take the mean for axes that will not appear in converted data
                if not silent and lims[1] - lims[0] > 0.001:
                    print(f"Data spans about {lims[1] - lims[0]:.3f} Å⁻¹ of {k}")
                other_coords[k] = np.array([(lims[0] + lims[1]) / 2])

        if not silent:
            print("Calculating destination coordinates")

        target_dict = self._inverse(
            (interp_coords | other_coords).get("kx"),
            (interp_coords | other_coords).get("ky"),
            (interp_coords | other_coords).get("kz", None),
        )

        if self.has_hv:
            old_dim_order = ("alpha", "hv")
            new_dim_order = (self.slit_axis, "kz")

        elif self.has_beta:
            old_dim_order = ("alpha", "beta")
            new_dim_order = ("ky", "kx")
        else:
            old_dim_order = ("alpha",)
            new_dim_order = (self.slit_axis,)

        if self.has_eV:
            old_dim_order = ("eV", *old_dim_order)
            new_dim_order = ("eV", *new_dim_order)
            new_size["eV"] = len(self._binding_energy)
            interp_coords["eV"] = np.squeeze(self._binding_energy)

        target = tuple(target_dict[dim] for dim in old_dim_order)

        for d in self._obj.dims:
            if d not in old_dim_order:
                old_dim_order += (d,)
                new_dim_order += (d,)
                new_size[d] = len(self._obj[d])
                interp_coords[d] = self._obj[d].values

        # Get non-dimension coordinates
        for d, c in self._obj.coords.items():
            if d not in interp_coords and d not in other_coords:
                if c.ndim == 0:
                    other_coords[d] = c.values

        # Only keep scalar coordinates
        for k, v in other_coords.items():
            if v.size != 1:
                del other_coords[k]
            else:
                other_coords[k] = v.item()

        if not silent:
            print(f"Converting {old_dim_order}  ->  {new_dim_order}")
            t_start = time.perf_counter()

        out = xr.DataArray(
            interpn(
                tuple(self._obj[dim].values for dim in old_dim_order),
                self._obj.copy(deep=True).transpose(*old_dim_order).values,
                target,
                bounds_error=False,
            ).reshape(tuple(new_size[d] for d in new_dim_order)),
            coords=interp_coords,
            dims=new_dim_order,
        )
        out = out.assign_attrs(self._obj.attrs)
        out = out.assign_coords(other_coords)
        if not silent:
            print(f"Interpolated in {time.perf_counter() - t_start:.3f} s")

        return out

    def interactive(self, **kwargs) -> ktool:
        """Open the interactive momentum space conversion tool."""
        if self._obj.ndim < 3:
            raise ValueError("Interactive tool requires three-dimensional data.")
        return ktool(self._obj, **kwargs)

    @only_momentum
    def hv_to_kz(self, hv: float) -> xr.DataArray:
        """Returns :math:`k_z` for a given photon energy.

        Useful when creating overlays on :math:`hν`-dependent data.

        Parameters
        ----------
        hv
            Photon energy in eV.

        Note
        ----
        This will be inexact for hv-dependent cuts that do not pass through the BZ
        center since we lost the exact angle values, i.e. the exact momentum
        perpendicular to the slit, during momentum conversion.
        """
        # Get kinetic energies for the given photon energy
        kinetic = hv - self.work_function + self._obj.eV

        # Get momentum conversion functions
        ang2k, k2ang = get_kconv_func(kinetic, self.configuration, self.angle_params)

        # Transformation yields in-plane momentum at given photon energy
        kx, ky = ang2k(*k2ang(self._obj.kx, self._obj.ky))

        # Calculate kz
        kz = kz_func(kinetic, self.inner_potential, kx, ky)

        return kz



__all__ = ["shift", "correct_with_edge"]

import itertools
from collections.abc import Callable

import lmfit
import matplotlib.pyplot as plt
import numpy as np
import numpy.typing as npt
import scipy.ndimage
import xarray as xr

from erlab.analysis.fit.models import FermiEdge2dModel
from erlab.plotting.colors import proportional_colorbar
from erlab.plotting.general import plot_array


def shift(
    darr: xr.DataArray,
    shift: float | xr.DataArray,
    along: str,
    shift_coords: bool = False,
    **shift_kwargs,
) -> xr.DataArray:
    """Shifts the values of a DataArray along a specified dimension.

    Parameters
    ----------
    darr
        The array to shift.
    shift
        The amount of shift to be applied along the specified dimension. If `shift` is a
        DataArray, different shifts can be applied to different coordinates. The
        dimensions of `shift` must be a subset of the dimensions of `darr`. For more
        information, see the note below. If `shift` is a `float`, the same shift is
        applied to all values along dimension `along`. This is equivalent to providing a
        0-dimensional DataArray.
    along
        Name of the dimension along which the shift is applied.
    shift_coords
        If `True`, the coordinates of the output data will be changed so that the output
        contains all the values of the original data. If `False`, the coordinates and
        shape of the original data will be retained, and only the data will be shifted.
        Defaults to `False`.
    **shift_kwargs
        Additional keyword arguments passed onto `scipy.ndimage.shift`. Default values
        of `cval` and `order` are set to `np.nan` and `1` respectively.

    Returns
    -------
    xarray.DataArray
        The shifted DataArray.

    Note
    ----
    - All dimensions in :code:`shift` must be a dimension in `darr`.
    - The :code:`shift` array values are divided by the step size along the `along`
      dimension.
    - NaN values in :code:`shift` are treated as zero.

    Example
    -------

    >>> import xarray as xr
    >>> darr = xr.DataArray(
    ...     np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]]).astype(float), dims=["x", "y"]
    ... )
    >>> shift_arr = xr.DataArray([1, 0, 2], dims=["x"])
    >>> shifted = erlab.analysis.utilities.shift(darr, shift_arr, along="y")
    >>> print(shifted)
    <xarray.DataArray (x: 3, y: 3)> Size: 72B
    nan 1.0 2.0 4.0 5.0 6.0 nan nan 7.0
    Dimensions without coordinates: x, y
    """

    shift_kwargs.setdefault("order", 1)
    shift_kwargs.setdefault("mode", "constant")
    if shift_kwargs["mode"] == "constant":
        shift_kwargs.setdefault("cval", np.nan)

    if not isinstance(shift, xr.DataArray):
        shift = xr.DataArray(float(shift))

    for dim in shift.dims:
        if dim not in darr.dims:
            raise ValueError(f"Dimension {dim} in shift array not found in input array")
        if darr[dim].size != shift[dim].size:
            raise ValueError(
                f"Dimension {dim} in shift array has different size than input array"
            )

    domain_indices: list[int] = [darr.get_axis_num(ax) for ax in shift.dims]

    # `along` must be evenly spaced and monotonic increasing
    out = darr.sortby(along).copy()

    # Normalize shift values
    along_step: float = out[along].values[1] - out[along].values[0]
    shift = (shift.copy() / along_step).fillna(0.0)

    if shift_coords:
        # We first apply the integer part of the average shift to the coords
        rigid_shift = np.round(shift.values.mean())
        shift = shift - rigid_shift

        # Apply coordinate shift
        out = out.assign_coords({along: out[along].values + rigid_shift * along_step})

        # The bounds of the remaining shift values are used to pad the data
        nshift_min, nshift_max = shift.values.min(), shift.values.max()
        pads: tuple[int] = min(0, round(nshift_min)), max(0, round(nshift_max))

        # Construct new coordinate array
        new_along = np.linspace(
            out[along].values[0] + pads[0] * along_step,
            out[along].values[-1] + pads[1] * along_step,
            out[along].size + sum(np.abs(pads)),
        )

        # Pad the data and assign new coordinates
        out = out.pad({along: np.abs(pads)}, mode="constant", constant_values=np.nan)
        out = out.assign_coords({along: new_along})

    for idxs in itertools.product(*[range(darr.shape[i]) for i in domain_indices]):
        # Construct slices for indexing
        slices = [slice(None)] * darr.ndim
        for domain_index, i in zip(domain_indices, idxs):
            slices[domain_index] = i
        slices = tuple(slices)

        # Initialize arguments to `scipy.ndimage.shift`
        input = out[slices]
        shifts = [0] * input.ndim
        shift_val: float = float(shift.isel(dict(zip(shift.dims, idxs))))
        shifts[input.get_axis_num(along)] = shift_val

        # Apply shift
        out[slices] = scipy.ndimage.shift(input.values, shifts, **shift_kwargs)

    return out


def correct_with_edge(
    darr: xr.DataArray,
    modelresult: lmfit.model.ModelResult | npt.NDArray[np.floating] | Callable,
    shift_coords: bool = True,
    plot: bool = False,
    plot_kw: dict | None = None,
    **shift_kwargs,
):
    """
    Corrects the given data array `darr` using the edge correction method.

    Parameters
    ----------
    darr
        The input data array to be corrected.
    modelresult
        The model result that contains the fermi edge information. It can be an instance
        of `lmfit.model.ModelResult`, a numpy array, or a callable function that takes
        an array of angles and returns the corresponding energy value.
    shift_coords
        Whether to shift the coordinates of the data array. Defaults to True.
    plot
        Whether to plot the original and corrected data arrays. Defaults to False.
    plot_kw
        Additional keyword arguments for the plot. Defaults to None.
    **shift_kwargs
        Additional keyword arguments to `shift`.

    Returns
    -------
    xarray.DataArray
        The edge corrected data.
    """
    if plot_kw is None:
        plot_kw = {}

    if isinstance(modelresult, lmfit.model.ModelResult):
        if isinstance(modelresult.model, FermiEdge2dModel):
            edge_quad = np.polynomial.polynomial.polyval(
                darr.alpha,
                np.array(
                    [
                        modelresult.best_values[f"c{i}"]
                        for i in range(modelresult.model.func.poly.degree + 1)
                    ]
                ),
            )
        else:
            edge_quad = modelresult.eval(x=darr.alpha)

    elif callable(modelresult):
        edge_quad = modelresult(darr.alpha.values)

    elif isinstance(modelresult, np.ndarray | xr.DataArray):
        if len(darr.alpha) != len(modelresult):
            raise ValueError(
                "Length of modelresult must be equal to the length of alpha in data"
            )
        else:
            edge_quad = modelresult

    else:
        raise ValueError(
            "modelresult must be one of "
            "lmfit.model.ModelResult, "
            "and np.ndarray or a callable"
        )

    if isinstance(edge_quad, np.ndarray):
        edge_quad = xr.DataArray(
            edge_quad, coords={"alpha": darr.alpha}, dims=["alpha"]
        )

    corrected = shift(darr, -edge_quad, "eV", shift_coords=shift_coords, **shift_kwargs)

    if plot is True:
        _, axes = plt.subplots(1, 2, layout="constrained", figsize=(10, 5))

        plot_kw.setdefault("cmap", "copper")
        plot_kw.setdefault("gamma", 0.5)

        if darr.ndim > 2:
            avg_dims = list(darr.dims)[:]
            avg_dims.remove("alpha")
            avg_dims.remove("eV")
            plot_array(darr.mean(avg_dims), ax=axes[0], **plot_kw)
            plot_array(corrected.mean(avg_dims), ax=axes[1], **plot_kw)
        else:
            plot_array(darr, ax=axes[0], **plot_kw)
            plot_array(corrected, ax=axes[1], **plot_kw)
        edge_quad.plot(ax=axes[0], ls="--", color="0.35")

        proportional_colorbar(ax=axes[0])
        proportional_colorbar(ax=axes[1])
        axes[0].set_title("Data")
        axes[1].set_title("Edge Corrected")

    return corrected



"""
Interactive plotting based on Qt and pyqtgraph

.. currentmodule:: erlab.interactive

Interactive tools
=================

.. autosummary::
   :toctree: generated

   imagetool
   bzplot
   colors
   curvefittingtool
   exampledata
   fermiedge
   kspace
   masktool
   derivative
   utilities

"""

__all__ = ["goldtool", "itool", "ktool", "dtool"]

import importlib

from erlab.interactive.fermiedge import goldtool
from erlab.interactive.imagetool import itool
from erlab.interactive.kspace import ktool

if importlib.util.find_spec("arpes"):
    from erlab.interactive.derivative import dtool



"""
Various functions for data analysis.

.. currentmodule:: erlab.analysis

Modules
=======

.. autosummary::
   :toctree: generated

   fit
   mask
   correlation
   gold
   interpolate
   kspace
   transform
   utilities

"""

__all__ = [
    "mask_with_polygon",
    "polygon_mask",
    "polygon_mask_points",
    "mask_with_hex_bz",
    "rotateinplane",
    "rotatestackinplane",
    "correct_with_edge",
]

from erlab.analysis import correlation, fit, gold, interpolate, kspace  # noqa: F401
from erlab.analysis.mask import (
    mask_with_hex_bz,
    mask_with_polygon,
    polygon_mask,
    polygon_mask_points,
)
from erlab.analysis.transform import rotateinplane, rotatestackinplane
from erlab.analysis.utilities import correct_with_edge



"""Functions for manipulating colors in Qt."""

from __future__ import annotations

__all__ = [
    "ColorMapComboBox",
    "ColorMapGammaWidget",
    "BetterImageItem",
    "BetterColorBarItem",
    "color_to_QColor",
    "pg_colormap_names",
    "pg_colormap_from_name",
    "pg_colormap_powernorm",
    "pg_colormap_to_QPixmap",
]

import weakref
from collections.abc import Iterable, Sequence
from typing import Literal

import matplotlib.colors as mcolors
import numpy as np
import numpy.typing as npt
import pyqtgraph as pg
from qtpy import QtCore, QtGui, QtWidgets


class ColorMapComboBox(QtWidgets.QComboBox):
    LOAD_ALL_TEXT = "Load all..."

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setPlaceholderText("Select colormap...")
        self.setToolTip("Colormap")
        w, h = 64, 16
        self.setIconSize(QtCore.QSize(w, h))
        # for name in pg_colormap_names("local"):
        for name in pg_colormap_names("all", exclude_local=True):
            self.addItem(name)
        # self.insertItem(0, self.LOAD_ALL_TEXT)
        self.thumbnails_loaded = False
        self.currentIndexChanged.connect(self.load_thumbnail)
        self.default_cmap = None

        sc_p = QtGui.QShortcut(QtGui.QKeySequence("Ctrl+Alt+Up"), self)
        sc_p.activated.connect(self.previousIndex)
        sc_m = QtGui.QShortcut(QtGui.QKeySequence("Ctrl+Alt+Down"), self)
        sc_m.activated.connect(self.nextIndex)

    def load_thumbnail(self, index: int):
        if not self.thumbnails_loaded:
            text = self.itemText(index)
            try:
                self.setItemIcon(index, QtGui.QIcon(pg_colormap_to_QPixmap(text)))
            except KeyError:
                pass

    def load_all(self):
        self.clear()
        for name in pg_colormap_names("all", exclude_local=True):
            self.addItem(QtGui.QIcon(pg_colormap_to_QPixmap(name)), name)
        self.resetCmap()
        self.showPopup()

    # https://forum.qt.io/topic/105012/qcombobox-specify-width-less-than-content/11
    def showPopup(self):
        maxWidth = self.maximumWidth()
        if maxWidth and maxWidth < 16777215:
            self.setPopupMinimumWidthForItems()
        if not self.thumbnails_loaded:
            for i in range(self.count()):
                self.load_thumbnail(i)
            self.thumbnails_loaded = True
        super().showPopup()

    def setPopupMinimumWidthForItems(self):
        view = self.view()
        fm = self.fontMetrics()
        maxWidth = max(
            [fm.boundingRect(self.itemText(i)).width() for i in range(self.count())]
        )
        if maxWidth:
            view.setMinimumWidth(maxWidth)

    @QtCore.Slot()
    def nextIndex(self):
        self.wheelEvent(
            QtGui.QWheelEvent(
                QtCore.QPoint(0, 0),
                QtCore.QPoint(0, 0),
                QtCore.QPoint(0, 0),
                QtCore.QPoint(0, -15),
                QtCore.Qt.MouseButton.NoButton,
                QtCore.Qt.KeyboardModifier.NoModifier,
                QtCore.Qt.ScrollPhase.ScrollUpdate,
                True,
            )
        )

    @QtCore.Slot()
    def previousIndex(self):
        self.wheelEvent(
            QtGui.QWheelEvent(
                QtCore.QPoint(0, 0),
                QtCore.QPoint(0, 0),
                QtCore.QPoint(0, 0),
                QtCore.QPoint(0, 15),
                QtCore.Qt.MouseButton.NoButton,
                QtCore.Qt.KeyboardModifier.NoModifier,
                QtCore.Qt.ScrollPhase.ScrollUpdate,
                True,
            )
        )

    def hidePopup(self):
        self.activated.emit(self.currentIndex())
        self.textActivated.emit(self.currentText())
        self.currentIndexChanged.emit(self.currentIndex())
        self.currentTextChanged.emit(self.currentText())
        super().hidePopup()

    def setDefaultCmap(self, cmap: str):
        self.default_cmap = cmap
        self.setCurrentText(cmap)

    def resetCmap(self):
        if self.default_cmap is None:
            self.setCurrentIndex(0)
        else:
            self.setCurrentText(self.default_cmap)


class ColorMapGammaWidget(QtWidgets.QWidget):
    valueChanged = QtCore.Signal(float)  #: :meta private:

    def __init__(
        self,
        parent: QtWidgets.QWidget = None,
        value: float = 1.0,
        slider_cls: type | None = None,
        spin_cls: type | None = None,
    ):
        super().__init__(parent=parent)
        self.setLayout(QtWidgets.QHBoxLayout(self))
        self.layout().setContentsMargins(0, 0, 0, 0)

        self.layout().setSpacing(3)

        if slider_cls is None:
            slider_cls = QtWidgets.QSlider
        if spin_cls is None:
            spin_cls = QtWidgets.QDoubleSpinBox

        self.spin = spin_cls(
            self,
            toolTip="Colormap gamma",
            decimals=2,
            wrapping=False,
            keyboardTracking=False,
            singleStep=0.1,
            minimum=0.01,
            maximum=99.99,
            value=value,
        )
        self.label = QtWidgets.QLabel("𝛾", self)
        self.label.setBuddy(self.spin)
        # self.label.setIndent(0)
        self.slider = slider_cls(
            self,
            toolTip="Colormap gamma",
            value=self.gamma_scale(value),
            singleStep=1,
            orientation=QtCore.Qt.Orientation.Horizontal,
        )
        self.slider.setMinimumWidth(75)
        self.spin.valueChanged.connect(self.spin_changed)

        self.slider.setRange(
            self.gamma_scale(self.spin.minimum()),
            self.gamma_scale(self.spin.maximum()),
        )
        self.slider.valueChanged.connect(self.slider_changed)

        self.layout().addWidget(self.label)
        self.layout().addWidget(self.spin)
        self.layout().addWidget(self.slider)

    def value(self) -> float:
        return self.spin.value()

    def setValue(self, value: float):
        self.spin.setValue(value)
        self.slider.setValue(self.gamma_scale(value))

    def spin_changed(self, value: float):
        self.slider.blockSignals(True)
        self.slider.setValue(self.gamma_scale(value))
        self.slider.blockSignals(False)
        self.valueChanged.emit(value)

    def slider_changed(self, value: float):
        self.spin.setValue(self.gamma_scale_inv(value))

    def gamma_scale(self, y: float) -> int:
        return round(1e4 * np.log10(y))

    def gamma_scale_inv(self, x: int) -> float:
        return np.power(10, x * 1e-4)


class BetterImageItem(pg.ImageItem):
    """:class:`pyqtgraph.ImageItem` with improved colormap support.

    Parameters
    ----------
    image
        Image data
    **kwargs
        Additional arguments to :class:`pyqtgraph.ImageItem`.

    Signals
    -------
    sigColorChanged()
    sigLimitChanged(float, float)

    """

    sigColorChanged = QtCore.Signal()  #: :meta private:

    def __init__(self, image: npt.NDArray = None, **kwargs):
        super().__init__(image, **kwargs)

    def set_colormap(
        self,
        cmap: pg.ColorMap | str,
        gamma: float,
        reverse: bool = False,
        highContrast: bool = False,
        zeroCentered: bool = False,
        update: bool = True,
    ):
        cmap = pg_colormap_powernorm(
            cmap,
            gamma,
            reverse,
            highContrast=highContrast,
            zeroCentered=zeroCentered,
        )
        self.set_pg_colormap(cmap, update=update)

    def set_pg_colormap(self, cmap: pg.ColorMap, update: bool = True):
        self._colorMap = cmap
        self.setLookupTable(cmap.getStops()[1], update=update)
        self.sigColorChanged.emit()


class BetterColorBarItem(pg.PlotItem):
    def __init__(
        self,
        parent: QtWidgets.QWidget | None = None,
        image: Sequence[BetterImageItem] | BetterImageItem | None = None,
        autoLevels: bool = False,
        limits: tuple[float, float] | None = None,
        pen: QtGui.QPen | str = "c",
        hoverPen: QtGui.QPen | str = "m",
        hoverBrush: QtGui.QBrush | str = "#FFFFFF33",
        **kargs,
    ):
        super().__init__(parent, **kargs)

        self.setDefaultPadding(0)
        # self.hideButtons()
        self.setMenuEnabled(False)
        self.vb.setMouseEnabled(x=False, y=True)

        self._colorbar = pg.ImageItem(
            np.linspace(0, 1, 4096).reshape((-1, 1)), axisOrder="row-major"
        )
        self.addItem(self._colorbar)

        self._span = pg.LinearRegionItem(
            (0, 1),
            "horizontal",
            swapMode="block",
            pen=pen,
            brush=pg.mkBrush(None),
            hoverPen=hoverPen,
            hoverBrush=hoverBrush,
        )
        self._span.setZValue(1000)
        self._span.lines[0].addMarker("<|>", size=6)
        self._span.lines[1].addMarker("<|>", size=6)

        self.addItem(self._span)

        self._fixedlimits: tuple[float, float] | None = None
        self.setAutoLevels(autoLevels)
        self._images: set[weakref.ref[BetterImageItem]] = set()
        self._primary_image: weakref.ref[BetterImageItem] | None = None
        if image is not None:
            self.setImageItem(image)
        if limits is not None:
            self.setLimits(limits)
        self.setLabels(right=("", ""))
        self.set_dimensions()

    @property
    def images(self):
        return self._images

    @property
    def primary_image(self):
        return self._primary_image

    @property
    def levels(self) -> Sequence[float]:
        return self.primary_image().getLevels()

    @property
    def limits(self) -> tuple[float, float]:
        if self._fixedlimits is not None:
            return self._fixedlimits
        else:
            return self.primary_image().quickMinMax(targetSize=2**16)

    def set_width(self, width: int):
        self.layout.setColumnFixedWidth(1, width)

    def set_dimensions(
        self,
        horiz_pad: int | None = None,
        vert_pad: int | None = None,
        font_size: float = 11.0,
    ):
        self.setSizePolicy(
            QtWidgets.QSizePolicy.Policy.Fixed, QtWidgets.QSizePolicy.Policy.Expanding
        )
        self.showAxes((True, True, True, True), showValues=(False, False, True, False))
        self.getAxis("top").setHeight(vert_pad)
        self.getAxis("bottom").setHeight(vert_pad)
        self.getAxis("right").setWidth(horiz_pad)
        self.getAxis("left").setWidth(None)

        font = QtGui.QFont()
        font.setPointSizeF(float(font_size))
        for axis in ("left", "bottom", "right", "top"):
            self.getAxis(axis).setTickFont(font)

    @QtCore.Slot()
    def level_change(self):
        if not self.isVisible():
            return
        for img_ref in self.images:
            img_ref().setLevels(self._span.getRegion())
        self.limit_changed()

    @QtCore.Slot()
    def level_change_fin(self):
        pass

    def setLimits(self, limits: tuple[float, float] | None):
        self._fixedlimits = limits
        if self._primary_image is not None:
            self.limit_changed()

    def addImage(self, image: Sequence[BetterImageItem] | BetterImageItem):
        # if isinstance(image, BetterImageItem):
        if not np.iterable(image):
            self._images.add(weakref.ref(image))
        else:
            for img in image:
                self._images.add(weakref.ref(img))

    def removeImage(self, image: Sequence[BetterImageItem] | BetterImageItem):
        if isinstance(image, Iterable):
            for img in image:
                self._images.remove(weakref.ref(img))
        else:
            self._images.remove(weakref.ref(image))

    def setImageItem(
        self,
        image: Sequence[BetterImageItem] | BetterImageItem,
        insert_in: pg.PlotItem | None = None,
    ):
        self.addImage(image)
        for img_ref in self._images:
            img = img_ref()
            if img is not None:
                if hasattr(img, "sigLevelsChanged"):
                    img.sigLevelsChanged.connect(self.image_level_changed)
                if hasattr(img, "sigImageChanged"):
                    img.sigImageChanged.connect(self.image_changed)
                if img.getColorMap() is not None:
                    self._primary_image = img_ref
                    break

        if self._primary_image is None:
            raise ValueError("ImageItem with a colormap was not found")
        # self.primary_image().sigLimitChanged.connect(self.limit_changed)

        # self.primary_image().sigImageChanged.connect(self.limit_changed)
        # self.primary_image().sigColorChanged.connect(self.color_changed)
        self.primary_image().sigColorChanged.connect(self.limit_changed)
        # else:
        # print("hello")

        if insert_in is not None:
            insert_in.layout.addItem(self, 2, 5)
            insert_in.layout.setColumnFixedWidth(4, 5)

        self._span.blockSignals(True)
        if hasattr(self, "limits"):
            self._span.setRegion(self.limits)
        self._span.blockSignals(False)

        self._span.sigRegionChanged.connect(self.level_change)
        self._span.sigRegionChangeFinished.connect(self.level_change_fin)
        self.image_changed()
        # self.color_changed()
        self.limit_changed()

    def image_level_changed(self):
        levels = self.primary_image().getLevels()
        if levels is not None:
            self._span.setRegion(levels)

    def image_changed(self):
        self.level_change()
        if self._auto_levels:
            self.reset_levels()

    def reset_levels(self):
        self._span.setRegion(self.limits)

    def setAutoLevels(self, value):
        self._auto_levels = bool(value)
        self._span.setVisible(not self._auto_levels)

        # self.isocurve.setParentItem(image)

    # def hideEvent(self, event: QtGui.QHideEvent):
    #     super().hideEvent(event)
    #     print("hide")

    # def showEvent(self, event: QtGui.QShowEvent):
    #     super().showEvent(event)
    #     # self._level_change()
    #     print("show")
    #     self.color_changed()
    #     self.limit_changed()

    # def setVisible(self, visible:bool, *args, **kwargs):
    # super().setVisible(visible, *args, **kwargs)
    # if visible:
    # self._level_change()
    # print('e')
    # self.isocurve.setVisible(visible, *args, **kwargs)

    def color_changed(self):
        if not self.isVisible():
            return
        cmap = self.primary_image()._colorMap
        lut = cmap.getStops()[1]
        if self._colorbar.image.shape[0] != lut.shape[0]:
            self._colorbar.setImage(cmap.pos.reshape((-1, 1)))
        self._colorbar._colorMap = cmap
        self._colorbar.setLookupTable(lut, update=True)

    # def limit_changed(self, mn: float | None = None, mx: float | None = None):
    def limit_changed(self):
        if not self.isVisible():
            return
        if not hasattr(self, "limits"):
            return
        self.color_changed()
        # if (self._fixedlimits is not None) or (mn is None):
        mn, mx = self.limits
        self._colorbar.setRect(0.0, mn, 1.0, mx - mn)
        if self.levels is not None:
            self._colorbar.setLevels((np.asarray(self.levels) - mn) / (mx - mn))
        self._span.setBounds((mn, mx))

    # def cmap_changed(self):
    #     cmap = self.imageItem()._colorMap
    #     # lut = self.imageItem().lut
    #     # lut = cmap.getLookupTable(nPts=4096)
    #     # lut = self._eff_lut_for_image(self.imageItem())
    #     # if lut is None:
    #         # lut = self.imageItem()._effectiveLut
    #     # if lut is not None:
    #         # print(lut)
    #     # if lut is None:
    #     lut = cmap.getStops()[1]
    #     # if not self.npts == lut.shape[0]:
    #     # self.npts = lut.shape[0]
    #     if not self._colorbar.image.shape[0] == lut.shape[0]:
    #         self._colorbar.setImage(cmap.pos.reshape((-1, 1)))
    #     self._colorbar._colorMap = cmap
    #     self._colorbar.setLookupTable(lut)
    #     # self._colorbar.setColorMap(cmap)

    def mouseDragEvent(self, ev):
        ev.ignore()


def color_to_QColor(
    c: str | tuple[float, ...], alpha: float | None = None
) -> QtGui.QColor:
    """Convert a matplotlib color to a :class:`PySide6.QtGui.QColor`.

    Parameters
    ----------
    c
        A valid matplotlib color. See the `matplotlib documentation
        <https://matplotlib.org/stable/tutorials/colors/colors.html>`_ for more
        information.
    alpha
        If supplied, applies transparency to the color.

    Returns
    -------
    PySide6.QtGui.QColor

    """
    return QtGui.QColor.fromRgbF(*mcolors.to_rgba(c, alpha=alpha))


def pg_colormap_names(
    source: Literal["local", "all", "matplotlib"] = "all", exclude_local: bool = False
) -> list[str]:
    """Get all valid :obj:`pyqtgraph` colormap names.

    Parameters
    ----------
    source
        If ``'all'``, includes all colorcet colormaps.

    Returns
    -------
    list of str

    """
    local = sorted(pg.colormap.listMaps())
    if source == "local":
        return local
    else:
        _mpl = sorted(pg.colormap.listMaps(source="matplotlib"))
        for cmap in _mpl:
            if cmap.startswith("cet_"):
                _mpl = list(filter((cmap).__ne__, _mpl))
            elif cmap.endswith("_r"):
                # _mpl_r.append(cmap)
                _mpl = list(filter((cmap).__ne__, _mpl))
        if source == "all":
            cet = sorted(pg.colormap.listMaps(source="colorcet"))
            # if (_mpl != []) and (cet != []):
            # local = []
            # _mpl_r = []

            if exclude_local:
                all_cmaps = cet + _mpl
            else:
                all_cmaps = local + cet + _mpl  # + _mpl_r
        elif exclude_local:
            all_cmaps = _mpl
        else:
            all_cmaps = local + _mpl
    return list(dict.fromkeys(all_cmaps))


def pg_colormap_from_name(name: str, skipCache: bool = True) -> pg.ColorMap:
    """Gets a :class:`pyqtgraph.ColorMap` from its name.

    Parameters
    ----------
    name
        A valid colormap name.
    skipCache
        Whether to skip cache, by default `True`. Passed onto
        :func:`pyqtgraph.colormap.get`.

    Returns
    -------
    pyqtgraph.ColorMap

    """
    try:
        return pg.colormap.get(name, skipCache=skipCache)
    except (FileNotFoundError, IndexError):
        try:
            return pg.colormap.get(name, source="matplotlib", skipCache=skipCache)
        except ValueError:
            return pg.colormap.get(name, source="colorcet", skipCache=skipCache)


def pg_colormap_powernorm(
    cmap: str | pg.ColorMap,
    gamma: float,
    reverse: bool = False,
    highContrast: bool = False,
    zeroCentered: bool = False,
    N: int = 65536,
) -> pg.ColorMap:
    if isinstance(cmap, str):
        cmap = pg_colormap_from_name(cmap, skipCache=True)

    if gamma == 1:

        def mapping_fn(x):
            return x

    elif highContrast:

        def mapping_fn(x):
            return 1 - np.power(np.flip(x), 1.0 / gamma)

    else:

        def mapping_fn(x):
            return np.power(x, gamma)

    x = np.linspace(0, 1, N)
    if zeroCentered:
        mapping = np.piecewise(
            x,
            [x < 0.5, x >= 0.5],
            [
                lambda x: 0.5 - 0.5 * mapping_fn(-2 * x + 1),
                lambda x: 0.5 + 0.5 * mapping_fn(2 * x - 1),
            ],
        )
    else:
        mapping = mapping_fn(x)
    mapping[mapping > 1] = 1
    mapping[mapping < 0] = 0

    if reverse:
        cmap.reverse()
    cmap.color = cmap.mapToFloat(mapping)
    cmap.pos = np.linspace(0, 1, N)
    pg.colormap._mapCache = {}  # disable cache to reduce memory usage
    return cmap


def pg_colormap_to_QPixmap(
    cmap: str | pg.ColorMap, w: int = 64, h: int = 16, skipCache: bool = True
) -> QtGui.QPixmap:
    """Converts a :class:`pyqtgraph.ColorMap` to a ``w``-by-``h`` QPixmap thumbnail.

    Parameters
    ----------
    cmap
        The colormap.
    w, h
        Specifies the dimension of the pixmap.
    skipCache : bool, optional
        Whether to skip cache, by default `True`. Passed onto
        :func:`pg_colormap_from_name`.

    Returns
    -------
    PySide6.QtGui.QPixmap

    """

    if isinstance(cmap, str):
        cmap = pg_colormap_from_name(cmap, skipCache=skipCache)
    # cmap_arr = np.reshape(cmap.getColors()[:, None], (1, -1, 4), order='C')
    # cmap_arr = np.reshape(
    # cmap.getLookupTable(0, 1, w, alpha=True)[:, None], (1, -1, 4),
    # order='C')
    cmap_arr = cmap.getLookupTable(0, 1, w, alpha=True)[:, None]

    # print(cmap_arr.shape)
    img = QtGui.QImage(cmap_arr, w, 1, QtGui.QImage.Format_RGBA8888)
    return QtGui.QPixmap.fromImage(img).scaled(w, h)



"""Transformations."""

__all__ = ["rotateinplane", "rotatestackinplane"]

import numpy as np
import xarray as xr


def rotateinplane(data: xr.DataArray, rotate, **interp_kwargs):
    interp_kwargs.setdefault("method", "linearfast")

    theta = np.radians(rotate)
    d0, d1 = data.dims
    x = xr.DataArray(data[d0] * np.cos(theta) - data[d1] * np.sin(theta))
    y = xr.DataArray(data[d0] * np.sin(theta) + data[d1] * np.cos(theta))
    return data.interp({d0: x, d1: y}, **interp_kwargs)


def rotatestackinplane(data: xr.DataArray, rotate, **interp_kwargs):
    interp_kwargs.setdefault("method", "linearfast")

    theta = np.radians(rotate)
    d0, d1, _ = data.dims
    x = xr.DataArray(data[d0] * np.cos(theta) - data[d1] * np.sin(theta))
    y = xr.DataArray(data[d0] * np.sin(theta) + data[d1] * np.cos(theta))
    return data.interp({d0: x, d1: y}, **interp_kwargs)



"""Data loader for beamline 4.0.3 at ALS."""

import glob
import os
import re
import datetime
import numpy as np
import numpy.typing as npt
import xarray as xr
import pandas as pd


import erlab.io.utilities
from erlab.io.dataloader import LoaderBase
from erlab.io.igor import load_experiment, load_wave


class BL403Loader(LoaderBase):
    name: str = "merlin"
    aliases: list[str] = ["ALS_BL4", "als_bl4", "BL403", "bl403"]

    name_map: dict[str, str | list[str]] = {
        "alpha": "deg",
        "beta": ["Polar", "Polar Compens"],
        "delta": "Azimuth",
        "xi": "Tilt",
        "x": "Sample X",
        "y": "Sample Y (Vert)",
        "z": "Sample Z",
        "hv": "BL Energy",
        "polarization": "EPU POL",
        "temp_sample": "Temperature Sensor B",
        "photon_flux": "Mesh Current",
    }
    coordinate_attrs: tuple[str, ...] = (
        "beta",
        "delta",
        "xi",
        "hv",
        "x",
        "y",
        "z",
        "polarization",
        "photon_flux",
    )
    additional_attrs: dict[str, str | int | float] = {
        "configuration": 1,
        "sample_workfunction": 4.44,
    }

    def load_single(self, file_path: str | os.PathLike) -> xr.DataArray:
        data = load_experiment(file_path)
        # One file always corresponds to single region, so assume only one data variable
        data: xr.DataArray = data.data_vars[next(iter(data.data_vars.keys()))]

        return self.process_keys(data)

    def identify(
        self, num: int, data_dir: str | os.PathLike
    ) -> tuple[list[str], dict[str, npt.NDArray[np.float64]]]:
        coord_dict: dict[str, npt.NDArray[np.float64]] = {}

        # Look for scans
        files = glob.glob(f"*_{str(num).zfill(3)}_S*.pxt", root_dir=data_dir)
        files.sort()
        # Assume files sorted by scan #

        if len(files) == 0:
            # Look for multiregion scan
            files = glob.glob(f"*_{str(num).zfill(3)}_R*.pxt", root_dir=data_dir)
            files.sort()
        elif len(files) > 1:
            prefix: str = re.match(
                r"(.*?)_" + str(num).zfill(3) + r"(?:_S\d{3})?.pxt", files[0]
            ).group(1)

            motor_file = os.path.join(
                data_dir, f"{prefix}_{str(num).zfill(3)}_Motor_Pos.txt"
            )

            coord_arr = np.loadtxt(motor_file, skiprows=1)
            with open(motor_file) as f:
                header = f.readline().strip().split("\t")

            if coord_arr.ndim == 1:
                coord_arr = coord_arr.reshape(-1, 1)

            for i, hdr in enumerate(header):
                key = self.name_map_reversed.get(hdr, hdr)
                coord_dict[key] = coord_arr[: len(files), i].astype(np.float64)

        if len(files) == 0:
            # Look for single file scan
            files = glob.glob(f"*_{str(num).zfill(3)}.pxt", root_dir=data_dir)

        if len(files) == 0:
            raise FileNotFoundError(f"No files found for scan {num} in {data_dir}")

        files = [os.path.join(data_dir, f) for f in files]

        return files, coord_dict

    def infer_index(self, name: str) -> int | None:
        try:
            scan_num: str = re.match(r".*?(\d{3})(?:_S\d{3})?", name).group(1)
        except (AttributeError, IndexError):
            return None

        if scan_num.isdigit():
            return int(scan_num)
        else:
            return None

    def post_process(self, data: xr.DataArray) -> xr.DataArray:
        data = super().post_process(data)

        if "eV" in data.coords:
            data = data.assign_coords(eV=-data.eV.values)

        return data

    def load_live(self, filename, data_dir=None):
        wave = load_wave(filename, data_dir)
        wave = wave.rename(
            {
                k: v
                for k, v in {"Energy": "eV", "Scienta": "alpha"}.items()
                if k in wave.dims
            }
        )
        wave = wave.assign_coords(eV=-wave["eV"] + wave.attrs["BL Energy"])

        return self.post_process(wave)

    def generate_summary(
        self, data_dir: str | os.PathLike, exclude_live: bool = False
    ) -> pd.DataFrame:
        files: dict[str, str] = {}

        for pth in erlab.io.utilities.get_files(data_dir, extensions=(".pxt",)):
            data_name = os.path.splitext(os.path.basename(pth))[0]
            name_match = re.match(r"(.*?_\d{3})_(?:_S\d{3})?", data_name)
            if name_match is not None:
                data_name = name_match.group(1)
            files[data_name] = pth

        if not exclude_live:
            for file in os.listdir(data_dir):
                if file.endswith(".ibw"):
                    data_name = os.path.splitext(file)[0]
                    path = os.path.join(data_dir, file)
                    files[data_name] = path

        summary_attrs: dict[str, str] = {
            "Lens Mode": "Lens Mode",
            "Scan Type": "Acquisition Mode",
            "T(K)": "temp_sample",
            "Pass E": "Pass Energy",
            "Analyzer Slit": "Slit Plate",
            "Polarization": "polarization",
            "hv": "hv",
            "Entrance Slit": "Entrance Slit",
            "Exit Slit": "Exit Slit",
            "x": "x",
            "y": "y",
            "z": "z",
            "polar": "beta",
            "tilt": "xi",
            "azi": "delta",
        }

        cols = ["Time", "File Name", "Type", *summary_attrs.keys()]

        data: list[dict] = []
        data_info = []

        for name, path in files.items():
            if os.path.splitext(path)[1] == ".ibw":
                data = self.load_live(path)
                if "beta" in data.dims:
                    data_type = "LP"
                else:
                    data_type = "LXY"
            else:
                data = self.load(path)
                if "beta" in data.dims:
                    data_type = "map"
                if "hv" in data.dims:
                    data_type = "hvdep"

            data_info.append(
                [
                    datetime.datetime.strptime(
                        f"{data.attrs['Date']} {data.attrs['Time']}",
                        "%d/%m/%Y %I:%M:%S %p",
                    ),
                    name,
                    data_type,
                ]
            )

            for k, v in summary_attrs.items():
                try:
                    val = data.attrs[v]
                except KeyError:
                    try:
                        val = data.coords[v].values
                        if val.size == 1:
                            val = val.item()
                    except KeyError:
                        val = ""

                if k == "Lens Mode":
                    val = val.replace("Angular", "A")

                elif k in ("Entrance Slit", "Exit Slit"):
                    val = round(val)

                elif k == "Polarization":
                    if np.iterable(val):
                        val = np.asarray(val).astype(int)
                    else:
                        val = [round(val)]
                    val = [{0: "LH", 2: "LV", -1: "RC", 1: "LC"}.get(v, v) for v in val]

                    if len(val) == 1:
                        val = val[0]

                data_info[-1].append(val)

            del data

        return (
            pd.DataFrame(data_info, columns=cols)
            .sort_values("Time")
            .set_index("File Name")
        )



"""Plugin for data acquired at KRISS."""

import os
import re

import numpy as np
import numpy.typing as npt

import erlab.io.utilities
from erlab.io.plugins.da30 import DA30Loader


class KRISSLoader(DA30Loader):
    name: str = "kriss"

    aliases: list[str] = ["KRISS"]

    coordinate_attrs: tuple[str, ...] = ("beta", "chi", "xi", "hv", "x", "y", "z")

    additional_attrs: dict[str, str | int | float] = {"configuration": 4}

    @property
    def name_map(self):
        return super().name_map | {"chi": "ThetaY", "xi": "ThetaX"}

    def identify(
        self, num: int, data_dir: str | os.PathLike
    ) -> tuple[list[str], dict[str, npt.NDArray[np.float64]]]:
        for file in erlab.io.utilities.get_files(data_dir, extensions=(".ibw", ".zip")):
            if file.endswith(".zip"):
                match = re.match(r"(.*?)" + str(num).zfill(4) + r".zip", file)
            else:
                match = re.match(
                    r"(.*?)" + str(num).zfill(4) + str(num).zfill(3) + r".ibw", file
                )

            if match is not None:
                return [file], {}

        raise FileNotFoundError(f"No files found for scan {num} in {data_dir}")



r"""Base functionality for implementing data loaders

This module provides a base class `LoaderBase` for implementing data loaders. Data
loaders are plugins used to load data from various file formats. Each data loader that
subclasses `LoaderBase` is registered on import in `loaders`.

Loaded ARPES data must contain several attributes and coordinates. See the
implementation of `LoaderBase.validate` for details.

For any data loader plugin subclassing `LoaderBase`, the following attributes and
methods must be defined: `LoaderBase.name`, `LoaderBase.aliases`,
`LoaderBase.name_map`, `LoaderBase.coordinate_attrs`, `LoaderBase.additional_attrs`,
`LoaderBase.always_single`, `LoaderBase.skip_validate`, :func:`LoaderBase.load_single`,
:func:`LoaderBase.identify`, :func:`LoaderBase.infer_index`, and
:func:`LoaderBase.generate_summary`.

If additional post-processing is required, the :func:`LoaderBase.post_process` method
can be extended to include the necessary functionality.

"""

from __future__ import annotations

import contextlib
import datetime
import itertools
import os
from typing import TYPE_CHECKING

import joblib
import numpy as np
import numpy.typing as npt
import pandas
import xarray as xr

if TYPE_CHECKING:
    from collections.abc import Iterable, Sequence


def _is_uniform(arr: npt.NDArray) -> bool:
    dif = np.diff(arr)
    return np.allclose(dif, dif[0], rtol=3e-05, atol=3e-05, equal_nan=True)


def _is_monotonic(arr: npt.NDArray) -> bool:
    dif = np.diff(arr)
    return np.all(dif >= 0) or np.all(dif <= 0)


class ValidationError(Exception):
    """This exception is raised when the loaded data fails validation checks."""


class LoaderBase:
    """Base class for all data loaders."""

    name: str | None = None
    """
    Name of the loader. Using a unique and descriptive name is recommended. For easy
    access, it is recommended to use a name that passes :func:`str.isidentifier`.
    """

    aliases: list[str] | None = None
    """List of alternative names for the loader."""

    name_map: dict[str, str | Iterable[str]] = {}
    """
    Dictionary that maps **new** coordinate or attribute names to **original**
    coordinate or attribute names. If there are multiple possible names for a single
    attribute, the value can be passed as an iterable.
    """

    coordinate_attrs: tuple[str, ...] = ()
    """
    Names of attributes (after renaming) that should be treated as coordinates.

    Note
    ----
    Although the data loader tries to preserve the original attributes, the attributes
    given here, both before and after renaming, will be removed from attrs for
    consistency.
    """

    additional_attrs: dict[str, str | int | float] = {}
    """Additional attributes to be added to the data."""

    always_single: bool = False
    """
    If `True`, this indicates that all individual scans always lead to a single data
    file. No concatenation of data from multiple files will be performed.
    """

    skip_validate: bool = False
    """If `True`, validation checks will be skipped."""

    @property
    def name_map_reversed(self) -> dict[str, str]:
        """A reversed version of the name_map dictionary."""
        return self.reverse_mapping(self.name_map)

    @staticmethod
    def reverse_mapping(mapping: dict[str, str | Iterable[str]]) -> dict[str, str]:
        """Reverse the given mapping dictionary to form a one-to-one mapping.

        Parameters
        ----------
        mapping
            The mapping dictionary to be reversed.

        Example
        -------

        >>> mapping = {"a": "1", "b": ["2", "3"]}
        >>> reverse_mapping(mapping)
        {'1': 'a', '2': 'b', '3': 'b'}

        """
        out: dict[str, str] = {}
        for k, v in mapping.items():
            if isinstance(v, str):
                out[v] = k
            else:
                for vi in v:
                    out[vi] = k
        return out

    @classmethod
    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)

        if not hasattr(cls, "name"):
            raise NotImplementedError("name attribute must be defined in the subclass")

        if not cls.name.startswith("_"):
            LoaderRegistry.instance().register(cls)

    @classmethod
    def formatter(cls, val: object):
        """Format the given value based on its type.

        This method is used when formtting the cells of the summary dataframe.

        Parameters
        ----------
        val
            The value to be formatted.

        Returns
        -------
        str or object
            The formatted value.

        Note
        ----
        This function formats the given value based on its type. It supports formatting
        for various types including numpy arrays, lists of strings, floating-point
        numbers, integers, and datetime objects.

        The function also tries to replace the Unicode hyphen-minus sign "-" (U+002D)
        with the better-looking Unicode minus sign "−" (U+2212) in most cases.

        - For numpy arrays:
            - If the array has a size of 1, the value is recursively formatted using
              `formatter(val.item())`.
            - If the array can be squeezed to a 1-dimensional array, the following are
              applied.

                - If the array is evenly spaced, the start, end, and step values are
                  formatted and returned as a string in the format "start→end (step)".
                - If the array is monotonic increasing or decreasing but not evenly
                  spaced, the start and end values are formatted and returned as a
                  string in the format "start→end".
                - If all elements are equal, the value is recursively formatted using
                  `formatter(val[0])`.
                - If the array is not monotonic, the mean and standard deviation values
                  are formatted and returned as a string in the format "mean±std".

            - For other dimensions, the array is returned as is.

        - For lists:
            The list is grouped by consecutive equal elements, and the count of each
            element is formatted and returned as a string in the format
            "[element]×count".

        - For floating-point numbers:
            - If the number is an integer, it is formatted as an integer using
              `formatter(np.int64(val))`.
            - Otherwise, it is formatted as a floating-point number with 4 decimal
              places and returned as a string.

        - For integers:
            The integer is returned as a string.

        - For datetime objects:
            The datetime object is formatted as a string in the format "%Y-%m-%d
            %H:%M:%S".

        - For other types:
            The value is returned as is.

        Examples
        --------

        >>> formatter(np.array([0.1, 0.15, 0.2]))
        '0.1→0.2 (0.05)'

        >>> formatter(np.array([1.0, 2.0, 2.1]))
        '1→2.1'

        >>> formatter(np.array([1.0, 2.1, 2.0]))
        '1.7±0.4967'

        >>> formatter([1, 1, 2, 2, 2, 3, 3, 3, 3])
        '[1]×2, [2]×3, [3]×4'

        >>> formatter(3.14159)
        '3.1416'

        >>> formatter(42)
        '42'

        >>> formatter(datetime.datetime(2024, 1, 1, 12, 0, 0, 0))
        '2024-01-01 12:00:00'
        """
        if isinstance(val, np.ndarray):
            if val.size == 1:
                return cls.formatter(val.item())

            elif val.squeeze().ndim == 1:
                val = val.squeeze()

                if _is_uniform(val):
                    start, end, step = tuple(
                        cls.formatter(v) for v in (val[0], val[-1], val[1] - val[0])
                    )
                    return f"{start}→{end} ({step})".replace("-", "−")

                elif _is_monotonic(val):
                    if val[0] == val[-1]:
                        return cls.formatter(val[0])

                    return f"{cls.formatter(val[0])}→{cls.formatter(val[-1])}"

                else:
                    return f"{cls.formatter(np.mean(val))}±{cls.formatter(np.std(val))}"

            else:
                return val

        elif isinstance(val, list):
            return ", ".join(
                [f"[{k}]×{len(tuple(g))}" for k, g in itertools.groupby(val)]
            )

        elif np.issubdtype(type(val), np.floating):
            if val.is_integer():
                return cls.formatter(np.int64(val))
            else:
                return np.format_float_positional(val, precision=4, trim="-").replace(
                    "-", "−"
                )
        elif np.issubdtype(type(val), np.integer):
            return str(val).replace("-", "−")

        elif isinstance(val, datetime.datetime):
            return val.strftime("%Y-%m-%d %H:%M:%S")

        else:
            return val

    @classmethod
    def get_styler(cls, df: pandas.DataFrame) -> pandas.io.formats.style.Styler:
        """Return a styled version of the given dataframe.

        This method, along with `formatter`, determines the display formatting of the
        summary dataframe. Override this method to change the display style.

        Parameters
        ----------
        df
            Summary dataframe as returned by `generate_summary`.

        Returns
        -------
        pandas.io.formats.style.Styler
            The styler to be displayed.

        """
        style = df.style.format(cls.formatter)
        if "Time" in df.columns:
            style = style.hide(["Time"], axis="columns")
        return style

    def load(
        self,
        identifier: str | os.PathLike | int | None,
        data_dir: str | os.PathLike | None = None,
        **kwargs: dict,
    ) -> xr.DataArray | xr.Dataset | list[xr.DataArray]:
        """Load ARPES data.

        Parameters
        ----------
        identifier
            Value that identifies a scan uniquely. If a string or path-like object is
            given, it is assumed to be the path to the data file. If an integer is
            given, it is assumed to be a number that specifies the scan number, and is
            used to automatically determine the path to the data file(s).
        data_dir
            Where to look for the data. If `None`, the default data directory will be
            used.
        single
            For some endstations, data for a single scan is saved over multiple files.
            This argument only has effect for such endstations. When a single file
            within a multiple file scan is provided to `identifier`, the default
            behavior when `single` is `False` is to infer the coordinates and return a
            single concatenated array that contains data from all files. If `single` is
            set to `True`, only the data from the file given is returned.
        **kwargs
            Additional keyword arguments are passed to `identify`.

        Returns
        -------
        xarray.DataArray or xarray.Dataset or list of xarray.DataArray
            The loaded data.

        """
        single = kwargs.pop("single", False)

        if self.always_single:
            single = True

        if isinstance(identifier, int):
            if data_dir is None:
                raise ValueError(
                    "data_dir must be specified when identifier is an integer"
                )
            file_paths, coord_dict = self.identify(identifier, data_dir, **kwargs)

            if len(file_paths) == 0:
                raise ValueError(
                    f"Failed to resolve identifier {identifier} "
                    f"for data directory {data_dir}"
                )
            elif len(file_paths) == 1:
                # Single file resolved
                data = self.load_single(file_paths[0])
            else:
                # Multiple files resolved
                data = self.combine_multiple(
                    self._load_multiple_parallel(file_paths), coord_dict
                )
        else:
            if data_dir is not None:
                # Generate full path to file
                identifier = os.path.join(data_dir, identifier)

            if not single:
                # Get file name without extension and path
                basename_no_ext: str = os.path.splitext(os.path.basename(identifier))[0]

                # Infer index from file name
                new_identifier: int | None = self.infer_index(basename_no_ext)

                if new_identifier is not None:
                    # On success, load with the index
                    new_dir: str = os.path.dirname(identifier)
                    return self.load(new_identifier, new_dir, single=single, **kwargs)
                else:
                    # On failure, assume single file
                    single = True

            data = self.load_single(identifier)

        data = self.post_process_general(data)

        if not self.skip_validate:
            self.validate(data)

        data.attrs["data_loader_name"] = str(self.name)

        return data

    def summarize(
        self, data_dir: str | os.PathLike, usecache: bool = True, **kwargs
    ) -> pandas.DataFrame:
        """
        Takes a path to a directory and summarizes the data in the directory to a
        `pandas.DataFrame`, much like a log file. This is useful for quickly inspecting
        the contents of a directory.

        The dataframe is formatted using `get_styler` and displayed in the IPython
        shell. Results are cached in a pickle file in the directory. If the pickle file
        is not found, the summary is generated with `generate_summary` and cached.

        Parameters
        ----------
        data_dir
            Directory to summarize.
        usecache
            Whether to use the cached summary if available. If `False`, the summary will
            be regenerated and cached.
        **kwargs
            Additional keyword arguments to be passed to `generate_summary`.

        Returns
        -------
        pandas.DataFrame
            Summary of the data in the directory.

        """
        pkl_path = os.path.join(data_dir, ".summary.pkl")
        df = None
        if usecache:
            try:
                df = pandas.read_pickle(pkl_path)
            except FileNotFoundError:
                pass

        df = df.head(len(df))

        if df is None:
            df = self.generate_summary(data_dir, **kwargs)
            df.to_pickle(pkl_path)

        try:
            shell = get_ipython().__class__.__name__  # type: ignore
            if shell in ["ZMQInteractiveShell", "TerminalInteractiveShell"]:
                from IPython.display import display

                with pandas.option_context(
                    "display.max_rows", len(df), "display.max_columns", len(df.columns)
                ):
                    display(self.get_styler(df))
        except NameError:
            pass

        return df

    def load_single(
        self, file_path: str | os.PathLike
    ) -> xr.DataArray | xr.Dataset | list[xr.DataArray]:
        r"""Load a single file and return it in applicable format.

        Any scan-specific postprocessing should be implemented in this method. When the
        single file contains many regions, the method should return a single dataset
        whenever the data can be merged with `xarray.merge` without conflicts.
        Otherwise, a list of `xarray.DataArray`\ s should be returned.

        Parameters
        ----------
        file_paths
            Full path to the file to be loaded.

        Returns
        -------
        xarray.DataArray or xarray.Dataset or list of xarray.DataArray
            The loaded data.

        """
        raise NotImplementedError("method must be implemented in the subclass")

    def identify(
        self, num: int, data_dir: str | os.PathLike
    ) -> tuple[list[str], dict[str, Iterable]]:
        """Identify the files and coordinates for a given scan number.

        This method takes a scan index and transforms it into a list of file paths and
        coordinates.

        Parameters
        ----------
        num
            The index of the scan to identify.
        data_dir
            The directory containing the data.

        Returns
        -------
        files : list[str]
            A list of file paths.
        coord_dict : dict[str, Iterable]
            A dictionary mapping scan axes names to scan coordinates. For scans spread
            over multiple files, the coordinates will be iterables corresponding to each
            file in the `files` list. For single file scans, an empty dictionary is
            returned.

        """
        raise NotImplementedError("method must be implemented in the subclass")

    def infer_index(self, name: str) -> int | None:
        """Infer the index for the given file name.

        Parameters
        ----------
        name
            The base name of the file without the path and extension.

        Returns
        -------
        index
            The inferred index if found, otherwise None.

        Note
        ----
        This method is used to determine all files for a given scan. Hence, for loaders
        with `always_single` set to `True`, this method does not have to be implemented.

        """
        raise NotImplementedError("method must be implemented in the subclass")

    def generate_summary(self, data_dir: str | os.PathLike) -> pandas.DataFrame:
        """Takes a path to a directory and summarizes the data in the directory to a
        pandas DataFrame, much like a log file. This is useful for quickly inspecting
        the contents of a directory.

        Parameters
        ----------
        data_dir
            Path to a directory.

        Returns
        -------
        pandas.DataFrame
            Summary of the data in the directory.

        """
        raise NotImplementedError("This loader does not support folder summaries")

    def combine_multiple(
        self,
        data_list: list[xr.DataArray | xr.Dataset],
        coord_dict: dict[str, Sequence],
    ) -> xr.DataArray | xr.Dataset | Sequence[xr.DataArray | xr.Dataset]:
        if len(coord_dict) == 0:
            try:
                # Try to merge the data without conflicts
                return xr.merge(data_list)
            except:  # noqa: E722
                # On failure, return a list
                return data_list
        else:
            for i in range(len(data_list)):
                data_list[i] = data_list[i].assign_coords(
                    {k: v[i] for k, v in coord_dict.items()}
                )
            return xr.concat(
                data_list, dim=next(iter(coord_dict.keys())), coords="different"
            )

    def post_process_general(
        self, data: xr.DataArray | xr.Dataset | list[xr.DataArray | xr.Dataset]
    ) -> xr.DataArray | xr.Dataset | list[xr.DataArray | xr.Dataset]:
        if isinstance(data, xr.DataArray):
            return self.post_process(data)

        elif isinstance(data, list):
            return [self.post_process(d) for d in data]

        elif isinstance(data, xr.Dataset):
            for k, v in data.data_vars.items():
                data[k] = self.post_process(v)
            return data

    def process_keys(
        self, data: xr.DataArray, key_mapping: dict[str, str] | None = None
    ) -> xr.DataArray:
        if key_mapping is None:
            key_mapping = self.name_map_reversed

        # Rename coordinates
        data = data.rename({k: v for k, v in key_mapping.items() if k in data.coords})

        # For attributes, keep original attribute and add new with renamed keys
        new_attrs = {}
        for k, v in dict(data.attrs).items():
            if k in key_mapping:
                new_key = key_mapping[k]
                if new_key in self.coordinate_attrs and new_key in data.coords:
                    # Renamed attribute is already a coordinate, remove
                    del data.attrs[k]
                else:
                    new_attrs[new_key] = v
        data = data.assign_attrs(new_attrs)

        # Move from attrs to coordinate if coordinate is not found
        data = data.assign_coords(
            {
                a: data.attrs.pop(a)
                for a in self.coordinate_attrs
                if a in data.attrs and a not in data.coords
            }
        )
        return data

    def post_process(self, data: xr.DataArray) -> xr.DataArray:
        data = self.process_keys(data)
        data = data.assign_attrs(self.additional_attrs)
        return data

    @classmethod
    def validate(cls, data: xr.DataArray | xr.Dataset):
        """Validate the input data to ensure it is in the correct format.

        Checks for the presence of all required coordinates and attributes. If the data
        does not pass validation, a `ValidationError` is raised. Validation is skipped
        for loaders with attribute `skip_validate` set to `True`.

        Parameters
        ----------
        data
            The data to be validated.

        Raises
        ------
        ValidationError

        """
        if isinstance(data, list):
            for d in data:
                cls.validate(d)
            return

        if isinstance(data, xr.Dataset):
            for v in data.data_vars.values():
                cls.validate(v)
            return

        for c in ("alpha", "beta", "delta", "xi", "hv"):
            if c not in data.coords:
                raise ValidationError(f"Missing coordinate {c}")
                # print(f"Missing coordinate {c}")

        for a in ("configuration", "temp_sample"):
            if a not in data.attrs:
                raise ValidationError(f"Missing attribute {c}")
                # print(f"Missing attribute {a}")

        if data.attrs["configuration"] not in (1, 2):
            if data.attrs["configuration"] not in (3, 4):
                raise ValidationError(
                    f"Invalid configuration {data.attrs['configuration']}"
                )
                # print(f"Invalid configuration {data.attrs['configuration']}")
            elif "chi" not in data.coords:
                raise ValidationError("Missing coordinate chi")
                # print("Missing coordinate chi")

    def _load_multiple_parallel(
        self, file_paths: list[str]
    ) -> list[xr.DataArray | xr.Dataset]:
        if len(file_paths) < 15:
            n_jobs = 1
        else:
            n_jobs = -1

        return joblib.Parallel(n_jobs=n_jobs)(
            joblib.delayed(self.load_single)(f) for f in file_paths
        )


class RegistryBase:
    """Base class for the loader registry.

    This class implements the singleton pattern, ensuring that only one instance of the
    registry is created and used throughout the application.
    """

    __instance = None

    def __new__(cls):
        if not isinstance(cls.__instance, cls):
            cls.__instance = super().__new__(cls)
        return cls.__instance

    @classmethod
    def instance(cls) -> LoaderRegistry:
        """Returns the registry instance."""
        return cls()


class LoaderRegistry(RegistryBase):
    loaders: dict[str, LoaderBase | type[LoaderBase]] = {}
    """Registered loaders \n\n:meta hide-value:"""

    alias_mapping: dict[str, str] = {}
    """Mapping of aliases to loader names \n\n:meta hide-value:"""

    current_loader: LoaderBase | None = None
    """Current loader \n\n:meta hide-value:"""

    default_data_dir: str | os.PathLike | None = None
    """Default directory to search for data files \n\n:meta hide-value:"""

    def register(self, loader_class: type[LoaderBase]):
        # Add class to loader
        self.loaders[loader_class.name] = loader_class

        # Add aliases to mapping
        self.alias_mapping[loader_class.name] = loader_class.name
        for alias in loader_class.aliases:
            self.alias_mapping[alias] = loader_class.name

    def get(self, key: str) -> LoaderBase:
        loader_name = self.alias_mapping.get(key)
        loader = self.loaders.get(loader_name)

        if loader is None:
            raise KeyError(f"Loader for {key} not found")

        if not isinstance(loader, LoaderBase):
            # If not an instance, create one
            self.loaders[loader_name] = loader()
            loader = self.loaders[loader_name]

        return loader

    def __getitem__(self, key: str) -> LoaderBase:
        return self.get(key)

    def __getattr__(self, key: str) -> LoaderBase:
        try:
            return self.get(key)
        except KeyError as e:
            raise AttributeError(f"Loader for {key} not found") from e

    def set_loader(self, loader: str | LoaderBase):
        """Set the current data loader.

        All subsequent calls to `load` will use the loader set here.

        Parameters
        ----------
        loader
            The loader to set. It can be either a string representing the name or alias
            of the loader, or a valid loader class.

        Example
        -------

        >>> erlab.io.set_loader("merlin")
        >>> dat_merlin_1 = erlab.io.load(...)
        >>> dat_merlin_2 = erlab.io.load(...)

        """
        if isinstance(loader, str):
            self.current_loader = self.get(loader)
        else:
            self.current_loader = loader

    @contextlib.contextmanager
    def loader_context(self, loader: str):
        """Context manager for temporarily changing the current data loader.

        The loader set here will only be used within the context manager.

        Parameters
        ----------
        loader
            The new loader to set as the current loader.

        Examples
        --------
        - Load data within a context manager:

          >>> with erlab.io.loader_context("merlin"):
          ...     dat_merlin = erlab.io.load(...)

        - Load data with different loaders:

          >>> erlab.io.set_loader("ssrl52")
          >>> dat_ssrl_1 = erlab.io.load(...)
          >>> with erlab.io.loader_context("merlin"):
          ...     dat_merlin = erlab.io.load(...)
          >>> dat_ssrl_2 = erlab.io.load(...)

        """
        old_loader: LoaderBase = self.current_loader
        self.set_loader(loader)
        try:
            yield self.current_loader
        finally:
            self.set_loader(old_loader)

    def set_data_dir(self, data_dir: str | os.PathLike):
        """Set the default data directory for the data loader.

        All subsequent calls to `load` will use the `data_dir` set here unless
        specified.

        Parameters
        ----------
        data_dir
            The path to a directory.

        Note
        ----
        This will only affect `load`. If the loader's ``load`` method is called
        directly, it will not use the default data directory.

        """
        self.default_data_dir = data_dir

    def load(
        self,
        identifier: str | os.PathLike | int | None,
        data_dir: str | os.PathLike | None = None,
        **kwargs: dict,
    ) -> xr.DataArray | xr.Dataset | list[xr.DataArray]:
        loader, default_dir = self._get_current_defaults()

        if data_dir is None:
            data_dir = default_dir

        if not isinstance(identifier, int) and os.path.isfile(identifier):
            # If the identifier is a path to a file, ignore data_dir
            data_dir = None

        return loader.load(identifier, data_dir=data_dir, **kwargs)

    def summarize(
        self, data_dir: str | os.PathLike | None = None, usecache: bool = True, **kwargs
    ) -> xr.DataArray | xr.Dataset | list[xr.DataArray]:
        loader, default_dir = self._get_current_defaults()

        if data_dir is None:
            data_dir = default_dir

        return loader.summarize(data_dir, usecache, **kwargs)

    def _get_current_defaults(self):
        if self.current_loader is None:
            raise ValueError(
                "No loader has been set. Set a loader with `erlab.io.set_loader` first"
            )
        return self.current_loader, self.default_data_dir

    def __repr__(self) -> str:
        out = "Registered data loaders\n=======================\n\n"
        out += "Loaders\n-------\n" + "\n".join(
            [f"{k}: {v}" for k, v in self.loaders.items()]
        )
        out += "\n\n"
        out += "Aliases\n-------\n" + "\n".join(
            [
                f"{k}: {v.aliases}"
                for k, v in self.loaders.items()
                if v.aliases is not None
            ]
        )
        return out

    def _repr_html_(self) -> str:
        out = ""
        out += "<table><thead>"
        out += (
            "<tr>"
            "<th style='text-align:left;'><b>Name</b></th>"
            "<th style='text-align:left;'><b>Aliases</b></th>"
            "<th style='text-align:left;'><b>Loader class</b></th>"
            "</tr>"
        )
        out += "</thead><tbody>"
        for k, v in self.loaders.items():
            aliases = ", ".join(v.aliases) if v.aliases is not None else ""

            # May be either a class or an instance
            if isinstance(v, LoaderBase):
                v = type(v)

            cls_name = f"{v.__module__}.{v.__qualname__}"

            out += (
                "<tr>"
                f"<td style='text-align:left;'>{k}</td>"
                f"<td style='text-align:left;'>{aliases}</td>"
                f"<td style='text-align:left;'>{cls_name}</td>"
                "</tr>"
            )
        out += "</tbody></table>"

        return out

    load.__doc__ = LoaderBase.load.__doc__
    summarize.__doc__ = LoaderBase.summarize.__doc__


loaders: LoaderRegistry = LoaderRegistry.instance()
"""Global instance of `LoaderRegistry`\n\n:meta hide-value:"""

set_loader = loaders.set_loader
loader_context = loaders.loader_context
set_data_dir = loaders.set_data_dir
load = loaders.load
summarize = loaders.summarize



"""Functions related to analyzing x-ray diffraction spectra.

Currently only supports loading raw data from igor ``.itx`` files.

"""

__all__ = ["load_xrd_itx"]

import ast
import re

import numpy as np
import xarray as xr


def load_xrd_itx(path: str, **kwargs: dict):
    r"""Load x-ray diffraction spectra from ``.itx`` file for Igor pro.

    Parameters
    ----------
    path
        Local path to ``.itx`` file.
    **kwargs
        Extra arguments to `open`.

    Returns
    -------
    xarray.Dataset
        Dataset object containing data from the file.

    Note
    ----
    By default, the file is read with the ``'windows-1252'`` encoding. This
    behavior can be customized by supplying keyword arguments.

    Examples
    --------
    Load from file:

    >>> xrd_data = load_xrd_itx("/path/to/example_data.itx")
    >>> xrd_data
    <xarray.Dataset>
    Dimensions:   (twotheta: 6701)
    Coordinates:
    * twotheta  (twotheta) float64 3.0 3.01 3.02 ... 69.98 69.99 70.0
    Data variables:
        yobs      (twotheta) float64 143.0 163.0 ... 7.0 7.0 7.0 2.0
        ycal      (twotheta) float64 119.4 118.8 ... 5.316 5.351 5.387
        bkg       (twotheta) float64 95.31 94.89 ... 5.228 5.264 5.3
        diff      (twotheta) float64 23.61 44.19 ... 1.684 1.649 -3.387

    Plot observed data:

    >>> xrd_data.yobs.plot()

    """
    kwargs.setdefault("encoding", "windows-1252")
    with open(path, **kwargs) as file:
        content = file.read()
    head, data = re.search(
        r"IGOR\nWAVES/O\s(.*?)\nBEGIN\n(.+?)\nEND", content, re.DOTALL
    ).groups()
    head = head.split(", ")

    data = np.array(
        ast.literal_eval("[[" + data.replace("\n", "],[").replace(" ", ",") + "]]")
    )
    ds = xr.Dataset({head[i]: ([head[0]], data[:, i]) for i in range(len(head))})
    if "diff" in ds.data_vars:
        ds = ds.rename_vars(diff="residual")
    return ds



"""Helper functions for fast slicing :class:`xarray.DataArray` objects."""

from __future__ import annotations

__all__ = ["ArraySlicer"]

import functools
from typing import TYPE_CHECKING

import numba
import numpy as np
import numpy.typing as npt
from qtpy import QtCore

from erlab.interactive.imagetool.fastbinning import _fast_nanmean_skipcheck

if TYPE_CHECKING:
    from collections.abc import Sequence

    import xarray as xr

VALID_NDIM = (2, 3, 4)

_signature_array_rect = [
    numba.types.UniTuple(numba.float32, 4)(
        numba.int64,
        numba.int64,
        numba.types.UniTuple(numba.types.UniTuple(numba.float32, 2), i),
        numba.types.UniTuple(numba.float32, i),
    )
    for i in VALID_NDIM
]
_signature_index_of_value = []
for ftype in (numba.float32, numba.float64):
    for i in VALID_NDIM:
        _signature_index_of_value.append(
            numba.int64(
                numba.int64,
                ftype,
                numba.types.UniTuple(numba.types.UniTuple(ftype, 2), i),
                numba.types.UniTuple(ftype, i),
                numba.types.UniTuple(numba.int64, i),
            )
        )


@numba.njit(_signature_array_rect, cache=True, fastmath=True)
def _array_rect(
    i: int,
    j: int,
    lims: tuple[tuple[np.float32, np.float32], ...],
    incs: tuple[np.float32, ...],
) -> tuple[np.float32, np.float32, np.float32, np.float32]:
    x = lims[i][0] - incs[i]
    y = lims[j][0] - incs[j]
    w = lims[i][-1] - x
    h = lims[j][-1] - y
    x += 0.5 * incs[i]
    y += 0.5 * incs[j]
    return x, y, w, h


@numba.njit(_signature_index_of_value, cache=True)
def _index_of_value(
    axis: int,
    val: np.floating,
    lims: tuple[tuple[np.floating, np.floating], ...],
    incs: tuple[np.floating, ...],
    shape: tuple[int],
) -> int:
    ind = min(
        round((val - lims[axis][0]) / incs[axis]),
        shape[axis] - 1,
    )
    if ind < 0:
        return 0
    return ind


@numba.njit(cache=True)
def _transposed(arr: npt.NDArray[np.floating]) -> npt.NDArray[np.floating]:
    if arr.ndim == 2:
        return arr.T
    elif arr.ndim == 3:
        return arr.transpose(1, 2, 0)
    else:
        return arr.transpose(1, 2, 3, 0)


@numba.njit(
    [numba.boolean(numba.float32[::1]), numba.boolean(numba.int64[::1])], cache=True
)
def _is_uniform(arr: npt.NDArray[np.float32]) -> bool:
    dif = np.diff(arr)
    return np.allclose(dif, dif[0], rtol=3e-05, atol=3e-05, equal_nan=True)


@numba.njit(
    [
        numba.int64(numba.float32[::1], numba.float32),
        numba.int64(numba.float64[::1], numba.float64),
    ],
    cache=True,
)
def _index_of_value_nonuniform(arr: npt.NDArray[np.float32], val: np.float32) -> int:
    return np.searchsorted((arr[:-1] + arr[1:]) / 2, val)


class ArraySlicer(QtCore.QObject):
    """Internal class used to slice a :class:`xarray.DataArray` rapidly.

    Computes binned line and image profiles from multiple cursors. Also handles the data
    indices and the number of bins for each cursor. Automatic conversion of non-uniform
    dimensions are also handled here.

    Parameters
    ----------
    xarray_obj
        A :class:`xarray.DataArray` with up to 4 dimensions.

    Signals
    -------
    sigIndexChanged(int, tuple)
    sigBinChanged(int, tuple)
    sigCursorCountChanged(int)
    sigShapeChanged()

    Note
    ----
    The original intent of this class was a xarray accessor. This is why `ArraySlicer`
    does not depend on a `ImageSlicerArea` but rather on the underlying
    `xarray.DataArray`. Originally, when loading a different array, a different instance
    of `ArraySlicer` had to be created. This was a terrible design choice since it
    messed up signals every time the instance was replaced. Hence, the behaviour was
    modified (23/06/19) so that the underlying `xarray.DataArray` of `ArraySlicer` could
    be swapped. As a consequence, each instance of `ImageSlicerArea` now corresponds to
    exactly one instance of `ArraySlicer`, regardless of the data. In the future,
    `ArraySlicer` might be changed so that it relies on its one-to-one correspondence
    with `ImageSlicerArea` for the signals.

    """

    sigIndexChanged = QtCore.Signal(int, object)  #: :meta private:
    sigBinChanged = QtCore.Signal(int, tuple)  #: :meta private:
    sigCursorCountChanged = QtCore.Signal(int)  #: :meta private:
    sigShapeChanged = QtCore.Signal()  #: :meta private:

    def __init__(self, xarray_obj: xr.DataArray):
        super().__init__()
        self._obj: xr.DataArray | None = None
        self.set_array(xarray_obj, validate=True, reset=True)

    @property
    def n_cursors(self) -> int:
        """The number of cursors."""
        return len(self._bins)

    @functools.cached_property
    def coords(self) -> tuple[npt.NDArray[np.float32], ...]:
        if self._nonuniform_axes:
            return tuple(
                (
                    self.values_of_dim(str(dim)[:-4])
                    if i in self._nonuniform_axes
                    else self.values_of_dim(dim)
                )
                for i, dim in enumerate(self._obj.dims)
            )
        else:
            return self.coords_uniform

    @functools.cached_property
    def coords_uniform(self) -> tuple[npt.NDArray[np.float32], ...]:
        return tuple(self.values_of_dim(dim) for dim in self._obj.dims)

    @functools.cached_property
    def incs(self) -> tuple[np.float32, ...]:
        return tuple(coord[1] - coord[0] for coord in self.coords)

    @functools.cached_property
    def incs_uniform(self) -> tuple[np.float32, ...]:
        return tuple(coord[1] - coord[0] for coord in self.coords_uniform)

    @functools.cached_property
    def lims(self) -> tuple[tuple[np.float32, np.float32], ...]:
        if self._nonuniform_axes:
            return tuple(
                (
                    (min(coord), max(coord))
                    if i in self._nonuniform_axes
                    else (coord[0], coord[-1])
                )
                for i, coord in enumerate(self.coords)
            )
        else:
            return tuple((coord[0], coord[-1]) for coord in self.coords)

    @functools.cached_property
    def lims_uniform(self) -> tuple[tuple[np.float32, np.float32], ...]:
        return tuple((coord[0], coord[-1]) for coord in self.coords_uniform)

    @functools.cached_property
    def data_vals_T(self) -> npt.NDArray[np.floating]:
        return _transposed(self._obj.values)

    # Benchmarks result in 10~20x slower speeds for bottleneck and numbagg compared to
    # numpy on arm64 mac with Accelerate BLAS. Needs confirmation on intel systems.
    @functools.cached_property
    def nanmax(self) -> np.floating:
        return np.nanmax(self._obj.values)

    @functools.cached_property
    def nanmin(self) -> np.floating:
        return np.nanmin(self._obj.values)

    @functools.cached_property
    def absnanmax(self) -> np.floating:
        return max(abs(self.nanmin), abs(self.nanmax))

    @functools.cached_property
    def absnanmin(self) -> np.floating:
        mn, mx = self.nanmin, self.nanmax
        if mn * mx <= np.float32(0.0):
            return np.float32(0.0)
        elif mn < np.float32(0.0):
            return -mx
        else:
            return mn

    @property
    def limits(self) -> tuple[np.floating, np.floating]:
        """Returns the global minima and maxima of the data."""
        return self.nanmin, self.nanmax

    @staticmethod
    def validate_array(data: xr.DataArray) -> xr.DataArray:
        """Validates a given :class:`xarray.DataArray`.

        If data has two momentum axes (``kx`` and ``ky``), set them (and ``eV`` if
        exists) as the first two (or three) dimensions. Then, checks the data for
        non-uniform coordinates, which are converted to indices. Finally, converts the
        coordinates to C-contiguous float32. If input data values neither float32 nor
        float64, a conversion to float64 is attempted.

        Parameters
        ----------
        data
            Input array with at least two dimensions.

        Returns
        -------
        xarray.DataArray
            The converted data.

        """
        # convert coords to C-contiguous float32
        data = data.assign_coords(
            {d: data[d].astype(np.float32, order="C") for d in data.dims}
        )

        if data.dtype not in (np.float32, np.float64):
            data = data.astype(np.float64)

        new_dims: tuple[str, ...] = ("kx", "ky")
        if all(d in data.dims for d in new_dims):
            # if data has kx and ky axis, transpose
            if "eV" in data.dims:
                new_dims += ("eV",)
            new_dims += tuple(d for d in data.dims if d not in new_dims)
            data = data.transpose(*new_dims)

        nonuniform_dims: list[str] = [
            str(d) for d in data.dims if not _is_uniform(data[d].values)
        ]
        for d in nonuniform_dims:
            data = data.assign_coords(
                {d + "_idx": (d, list(np.arange(len(data[d]), dtype=np.float32)))}
            ).swap_dims({d: d + "_idx"})

        return data

    def reset_property_cache(self, propname: str) -> None:
        self.__dict__.pop(propname, None)

    def clear_dim_cache(self):
        for prop in (
            "coords",
            "coords_uniform",
            "incs",
            "incs_uniform",
            "lims",
            "lims_uniform",
            "data_vals_T",
        ):
            self.reset_property_cache(prop)

    def clear_val_cache(self):
        for prop in ("nanmax", "nanmin", "absnanmax", "absnanmin"):
            self.reset_property_cache(prop)

    def clear_cache(self):
        self.clear_dim_cache()
        self.clear_val_cache()

    def set_array(
        self, xarray_obj: xr.DataArray, validate: bool = True, reset: bool = False
    ) -> None:
        del self._obj

        if validate:
            self._obj: xr.DataArray = self.validate_array(xarray_obj)
        else:
            self._obj: xr.DataArray = xarray_obj
        self._nonuniform_axes: list[str] = [
            i for i, d in enumerate(self._obj.dims) if str(d).endswith("_idx")
        ]

        self.clear_dim_cache()
        if validate:
            self.clear_val_cache()

        if reset:
            self._bins: list[list[int]] = [[1] * self._obj.ndim]
            self._indices: list[list[int]] = [
                [s // 2 - (1 if s % 2 == 0 else 0) for s in self._obj.shape]
            ]
            self._values: list[list[np.float32]] = [
                [c[i] for c, i in zip(self.coords, self._indices[0])]
            ]
            self.snap_to_data: bool = False

    def values_of_dim(self, dim: str) -> npt.NDArray[np.float32]:
        """Fast equivalent of :code:`self._obj[dim].values`.

        Returns the cached pointer of the underlying coordinate array, achieving a ~80x
        speedup. This should work most of the time since we only assume floating point
        values, but does require further testing. May break for future versions of
        :obj:`pandas` or :obj:`xarray`. See Notes.

        Parameters
        ----------
        dim
            Name of the dimension to get the values from.

        Returns
        -------
        numpy.ndarray

        Notes
        -----
        Looking at the implementation, I think this may return a pandas array in some
        cases, but I'm not sure so I'll just leave it this way. When something breaks,
        replacing with :code:`self._obj._coords[dim]._data.array.array._ndarray` may
        do the trick.

        """
        return self._obj._coords[dim]._data.array._data

    def add_cursor(self, like_cursor: int = -1, update: bool = True) -> None:
        self._bins.append(list(self.get_bins(like_cursor)))
        new_ind = self.get_indices(like_cursor)
        self._indices.append(list(new_ind))
        self._values.append([c[i] for c, i in zip(self.coords, new_ind)])
        if update:
            self.sigCursorCountChanged.emit(self.n_cursors)

    def remove_cursor(self, index: int, update: bool = True) -> None:
        if self.n_cursors == 1:
            raise ValueError("There must be at least one cursor.")
        self._bins.pop(index)
        self._indices.pop(index)
        self._values.pop(index)
        if update:
            self.sigCursorCountChanged.emit(self.n_cursors)

    def center_cursor(self, cursor: int, update: bool = True) -> None:
        self.set_indices(
            cursor,
            [s // 2 - (1 if s % 2 == 0 else 0) for s in self._obj.shape],
            update=update,
        )

    @QtCore.Slot(int, result=list)
    def get_bins(self, cursor: int) -> list[int]:
        return self._bins[cursor]

    def set_bins(
        self, cursor: int, value: list[int | None], update: bool = True
    ) -> None:
        if not len(value) == self._obj.ndim:
            raise ValueError("length of bin array must match the number of dimensions.")
        axes: list[int | None] = []
        for i, x in enumerate(value):
            axes += self.set_bin(cursor, i, x, update=False)
        if update:
            self.sigBinChanged.emit(cursor, tuple(axes))

    @QtCore.Slot(int, int, int, bool, result=list)
    def set_bin(
        self, cursor: int, axis: int, value: int, update: bool = True
    ) -> list[int | None]:
        if value is None:
            return []
        if int(value) != value:
            raise TypeError("bins must have integer type")
        self._bins[cursor][axis] = int(value)
        if update:
            self.sigBinChanged.emit(cursor, (axis,))
            return []
        return [axis]

    @QtCore.Slot(int, result=tuple)
    def get_binned(self, cursor: int) -> tuple[bool, ...]:
        return tuple(b != 1 for b in self.get_bins(cursor))

    @QtCore.Slot(int, result=list)
    def get_indices(self, cursor: int) -> list[int]:
        return self._indices[cursor]

    @QtCore.Slot(int, int, result=int)
    def get_index(self, cursor: int, axis: int) -> int:
        return self._indices[cursor][axis]

    def set_indices(self, cursor: int, value: list[int], update: bool = True) -> None:
        if not len(value) == self._obj.ndim:
            raise ValueError(
                "length of index array must match the number of dimensions"
            )
        axes: list[int | None] = []
        for i, x in enumerate(value):
            axes += self.set_index(cursor, i, x, update=False)
        if update:
            self.sigIndexChanged.emit(cursor, tuple(axes))

    @QtCore.Slot(int, int, int, bool, result=list)
    def set_index(
        self, cursor: int, axis: int, value: int, update: bool = True
    ) -> list[int | None]:
        if value is None:
            return []
        if int(value) != value:
            raise TypeError("indices must have integer type")
        self._indices[cursor][axis] = int(value)
        self._values[cursor][axis] = self.coords[axis][int(value)]
        if update:
            self.sigIndexChanged.emit(cursor, (axis,))
            return []
        return [axis]

    @QtCore.Slot(int, int, int, bool)
    def step_index(
        self, cursor: int, axis: int, value: int, update: bool = True
    ) -> None:
        self._indices[cursor][axis] += value
        if (
            self._indices[cursor][axis] >= self._obj.shape[axis]
            or self._indices[cursor][axis] < 0
        ):
            self._indices[cursor][axis] -= value
            return
        self._values[cursor][axis] = self.coords[axis][self._indices[cursor][axis]]
        if update:
            self.sigIndexChanged.emit(cursor, (axis,))

    @QtCore.Slot(int, bool, result=list)
    def get_values(self, cursor: int, uniform: bool = False) -> list[np.float32]:
        if uniform and self._nonuniform_axes:
            val = list(self._values[cursor])
            for ax in self._nonuniform_axes:
                val[ax] = np.float32(self._indices[cursor][ax])
            return val
        else:
            return self._values[cursor]

    @QtCore.Slot(int, int, bool, result=np.float32)
    def get_value(self, cursor: int, axis: int, uniform: bool = False) -> np.float32:
        if uniform and axis in self._nonuniform_axes:
            return np.float32(self._indices[cursor][axis])
        else:
            return self._values[cursor][axis]

    def set_values(
        self, cursor: int, value: list[np.float32], update: bool = True
    ) -> None:
        if not len(value) == self._obj.ndim:
            raise ValueError(
                "length of value array must match the number of dimensions"
            )
        axes: list[int | None] = []
        for i, x in enumerate(value):
            axes += self.set_value(cursor, i, x, update=False)
        if update:
            self.sigIndexChanged.emit(cursor, tuple(axes))

    @QtCore.Slot(int, int, np.float32, bool, bool, result=list)
    def set_value(
        self,
        cursor: int,
        axis: int,
        value: np.float32,
        update: bool = True,
        uniform: bool = False,
    ) -> list[int | None]:
        if value is None:
            return []
        self._indices[cursor][axis] = self.index_of_value(axis, value, uniform=uniform)
        if self.snap_to_data or (axis in self._nonuniform_axes):
            new = self.coords[axis][self._indices[cursor][axis]]
            if self._values[cursor][axis] == new:
                return []
            self._values[cursor][axis] = new
        else:
            self._values[cursor][axis] = np.float32(value)
        if update:
            self.sigIndexChanged.emit(cursor, (axis,))
            return []
        return [axis]

    @QtCore.Slot(int, bool, result=np.floating)
    def point_value(
        self, cursor: int, binned: bool = True
    ) -> npt.NDArray[np.floating] | np.floating:
        if binned:
            return self.extract_avg_slice(cursor, tuple(range(self._obj.ndim)))
        else:
            return self._obj.values[tuple(self.get_indices(cursor))]

    @QtCore.Slot(int, int)
    def swap_axes(self, ax1: int, ax2: int) -> None:
        for i in range(self.n_cursors):
            self._bins[i][ax1], self._bins[i][ax2] = (
                self._bins[i][ax2],
                self._bins[i][ax1],
            )
            self._values[i][ax1], self._values[i][ax2] = (
                self._values[i][ax2],
                self._values[i][ax1],
            )
            self._indices[i][ax1], self._indices[i][ax2] = (
                self._indices[i][ax2],
                self._indices[i][ax1],
            )
        dims_new = list(self._obj.dims)
        dims_new[ax1], dims_new[ax2] = dims_new[ax2], dims_new[ax1]

        self.set_array(self._obj.transpose(*dims_new), validate=False)

        self.sigShapeChanged.emit()

    def array_rect(
        self, i: int | None = None, j: int | None = None
    ) -> (
        tuple[np.float32, np.float32, np.float32, np.float32] | npt.NDArray[np.float32]
    ):
        if i is None:
            i = 0
        if j is None:
            return self.coords_uniform[i]
        return _array_rect(i, j, self.lims_uniform, self.incs_uniform)

    def value_of_index(
        self, axis: int, value: int, uniform: bool = False
    ) -> np.float32:
        if uniform or (axis not in self._nonuniform_axes):
            return self.coords_uniform[axis][value]
        else:
            return self.coords[axis][value]

    def index_of_value(
        self, axis: int, value: np.float32, uniform: bool = False
    ) -> int:
        if uniform or (axis not in self._nonuniform_axes):
            return _index_of_value(
                axis, value, self.lims_uniform, self.incs_uniform, self._obj.shape
            )
        else:
            return _index_of_value_nonuniform(self.coords[axis], value)

    def gen_selection_code(self, cursor: int, disp: Sequence[int]) -> str:
        axis = sorted(set(range(self._obj.ndim)) - set(disp))
        slice_dict = {
            self._obj.dims[ax]: self._bin_slice(cursor, ax, int_if_one=True)
            for ax in axis
        }
        return f".isel(**{slice_dict!s}).squeeze()"

    def xslice(self, cursor: int, disp: Sequence[int]) -> xr.DataArray:
        axis = sorted(set(range(self._obj.ndim)) - set(disp))
        slices = {self._obj.dims[ax]: self._bin_slice(cursor, ax) for ax in axis}
        return self._obj.isel(**slices).squeeze()

    @QtCore.Slot(int, tuple, result=np.ndarray)
    def slice_with_coord(
        self, cursor: int, disp: Sequence[int]
    ) -> tuple[
        tuple[np.float32, np.float32, np.float32, np.float32] | npt.NDArray[np.float32],
        npt.NDArray[np.float32] | np.float32,
    ]:
        axis = sorted(set(range(self._obj.ndim)) - set(disp))
        return self.array_rect(*disp), self.extract_avg_slice(cursor, axis)

    def extract_avg_slice(
        self, cursor: int, axis: Sequence[int]
    ) -> npt.NDArray[np.floating] | np.floating:
        if len(axis) == 0:
            return self.data_vals_T
        elif len(axis) == 1:
            return self._bin_along_axis(cursor, axis[0])
        else:
            return self._bin_along_multiaxis(cursor, axis)

    def span_bounds(self, cursor: int, axis: int) -> npt.NDArray[np.float32]:
        slc = self._bin_slice(cursor, axis)
        lb = max(0, slc.start)
        ub = min(self._obj.shape[axis] - 1, slc.stop - 1)
        return self.coords_uniform[axis][[lb, ub]]

    def _bin_slice(
        self, cursor: int, axis: int, int_if_one: bool = False
    ) -> slice | int:
        center = self.get_indices(cursor)[axis]
        if self.get_binned(cursor)[axis]:
            window = self.get_bins(cursor)[axis]
            return slice(center - window // 2, center + (window - 1) // 2 + 1)
        elif int_if_one:
            return center
        else:
            return slice(center, center + 1)

    def _bin_along_axis(
        self, cursor: int, axis: int
    ) -> npt.NDArray[np.floating] | np.floating:
        axis_val = (axis - 1) % self._obj.ndim
        if not self.get_binned(cursor)[axis]:
            return self.data_vals_T[
                (slice(None),) * axis_val + (self._bin_slice(cursor, axis),)
            ].squeeze(axis=axis_val)
        else:
            return _fast_nanmean_skipcheck(
                self.data_vals_T[
                    (slice(None),) * axis_val + (self._bin_slice(cursor, axis),)
                ],
                axis=axis_val,
            )

    def _bin_along_multiaxis(
        self, cursor: int, axis: Sequence[int]
    ) -> npt.NDArray[np.floating] | np.floating:
        if any(self.get_binned(cursor)):
            slices = tuple(self._bin_slice(cursor, ax) for ax in axis)
        else:
            slices = tuple(self.get_indices(cursor)[i] for i in axis)
        axis = tuple((ax - 1) % self._obj.ndim for ax in axis)
        selected = self.data_vals_T[
            tuple(
                slices[axis.index(d)] if d in axis else slice(None)
                for d in range(self._obj.ndim)
            )
        ]
        if any(self.get_binned(cursor)):
            return _fast_nanmean_skipcheck(selected, axis=axis)
        else:
            return selected



from __future__ import annotations

from collections.abc import Iterable, Sequence
from typing import TYPE_CHECKING

import iminuit.cost
import iminuit.util
import matplotlib.pyplot as plt
import numpy as np
import numpy.typing as npt
from iminuit.util import _detect_log_spacing, _smart_sampling

import erlab.plotting.general

if TYPE_CHECKING:
    import lmfit


class LeastSq(iminuit.cost.LeastSquares):
    """A thin wrapper around `iminuit.cost.LeastSquares` that produces better plots."""

    def visualize(
        self, args: npt.ArrayLike, model_points: int | Sequence[float] = 0
    ) -> tuple[
        tuple[npt.NDArray, npt.NDArray, npt.NDArray], tuple[npt.NDArray, npt.NDArray]
    ]:
        if self._ndim > 1:
            raise ValueError("visualize is not implemented for multi-dimensional data")

        plt.grid(visible="both")
        x, y, ye = self._masked.T
        plt.errorbar(
            x, y, ye, fmt="o", lw=0.75, ms=3, mfc="w", zorder=2, c="0.4", capsize=0
        )
        if isinstance(model_points, Iterable):
            xm = np.array(model_points)
            ym = self.model(xm, *args)
        elif model_points > 0:
            if _detect_log_spacing(x):
                xm = np.geomspace(x[0], x[-1], model_points)
            else:
                xm = np.linspace(x[0], x[-1], model_points)
            ym = self.model(xm, *args)
        else:
            xm, ym = _smart_sampling(
                lambda x: self.model(x, *args), x[0], x[-1], start=len(x)
            )
        plt.plot(xm, ym, "r-", lw=1, zorder=3)
        return (x, y, ye), (xm, ym)

    visualize.__doc__ = iminuit.cost.LeastSquares.visualize.__doc__


class _TempFig:
    def __init__(self, w, h):
        self.fig = plt.figure(figsize=(w, h), layout="constrained")

    def __enter__(self) -> None:
        pass

    def __exit__(self, *args: object) -> None:
        plt.close(self.fig)


class Minuit(iminuit.Minuit):
    """A subclass of `iminuit.Minuit` that provides additional functionality for fitting
    `lmfit` library models.

    This class extends the functionality of the `iminuit.Minuit` class by providing a
    convenient method `from_lmfit` to initialize the `Minuit` object from an
    `lmfit.Model` object.

    For more information on the `iminuit` library, see its documentation.

    Examples
    --------

    >>> import lmfit.models
    >>> import numpy as np
    >>> from erlab.analysis.fit.minuit import Minuit

    >>> # Create an lmfit.Model object
    >>> model = lmfit.models.LinearModel()

    >>> # Generate some data
    >>> x = np.linspace(0, 10, 100)
    >>> y = model.eval(x=x, a=2, b=1)
    >>> rng = np.random.default_rng(1)
    >>> y = rng.normal(y, 0.5)

    >>> # Initialize a Minuit object from the lmfit.Model object
    >>> m = Minuit.from_lmfit(model, y, x)

    >>> # Perform the fit
    >>> m.migrad()

    """

    @classmethod
    def from_lmfit(
        cls,
        model: lmfit.Model,
        data: npt.ArrayLike,
        ivars: npt.ArrayLike | Sequence[npt.ArrayLike],
        yerr: float | npt.ArrayLike | None = None,
        return_cost: bool = False,
        **kwargs,
    ) -> Minuit | tuple[LeastSq, Minuit]:
        if len(model.independent_vars) == 1:
            if len(ivars) != 1:
                ivars = [ivars]

        x = [np.asarray(a) for a in ivars]

        if len(x) != len(model.independent_vars):
            raise ValueError("Number of independent variables does not match model.")

        if len(x) == 1:
            x = x[0]

        if yerr is None:
            yerr = 1.0

        try:
            if len(model.independent_vars) == 1:
                params = model.guess(data, x)
            else:
                params = model.guess(data, *x)

            for key, val in kwargs.items():
                pname = f"{model.prefix}{key}"
                if pname not in params:
                    pname = key
                if pname not in params:
                    continue
                if isinstance(val, dict):
                    params[pname].set(**val)
                else:
                    params[pname].value = val

        except NotImplementedError:
            params = model.make_params(**kwargs)

        # Convert data to numpy array (must be after guessing parameters)
        data = np.asarray(data)

        fixedparams = []
        values: dict[str, float] = {}
        limits: dict[str, tuple[float, float]] = {}

        for k, par in params.items():
            if not par.vary:
                fixedparams.append(k)

            val = float(par.value)
            if not np.isfinite(val):
                val = 0.0

            values[k] = val
            limits[k] = (float(par.min), float(par.max))

        # Convert to kwargs
        if len(model.independent_vars) == 1:

            def _temp_func(x, *fargs):
                return model.func(x, **dict(zip(model._param_root_names, fargs)))

        else:

            def _temp_func(x, *fargs):
                return model.func(*x, **dict(zip(model._param_root_names, fargs)))

        c = LeastSq(x, data, yerr, _temp_func)
        m = cls(c, name=model.param_names, **values)

        for n in model.param_names:
            m.fixed[n] = n in fixedparams
            m.limits[n] = limits[n]

        if return_cost:
            return c, m
        return m

    def _repr_html_(self):
        s = ""
        if self.fmin is not None:
            s += self.fmin._repr_html_()
        s += self.params._repr_html_()
        if self.merrors:
            s += self.merrors._repr_html_()
        if self.covariance is not None:
            s += self.covariance._repr_html_()
        if self.fmin is not None:
            try:
                self.visualize()

                with _TempFig(*erlab.plotting.general.figwh()):
                    self.visualize()
                    # import io
                    # with io.StringIO() as io:
                    # plt.savefig(io, format="svg", dpi=10)
                    # io.seek(0)
                    # s += io.read()
            except (ModuleNotFoundError, AttributeError, ValueError):
                pass
        return s



import xarray as xr

try:
    import csaps
except ImportError as e:
    raise ImportError(
        "The `csaps` package is required for this module. "
        "Please install it using `pip install csaps`."
    ) from e


def xcsaps(
    arr: xr.DataArray, **kwargs: dict
) -> tuple[xr.DataArray, csaps.ISmoothingSpline]:
    """`xarray` compatible `csaps.csaps`.

    Parameters
    ----------
    arr
        Input array for smoothing spline calculation.
    **kwargs
        Keyword arguments for :func:`csaps.csaps`. `normalizedsmooth` is set to `True`
        by default.

    Returns
    -------
    out : xarray.DataArray
        Smoothing spline evaluated at `arr` coordinates.
    spl : csaps.ISmoothingSpline
        The spline object.
    """

    kwargs.setdefault("normalizedsmooth", True)
    coords = [arr[d].values for d in arr.dims]
    spl = csaps.csaps(coords, arr.values, **kwargs)
    out = xr.DataArray(
        data=spl(coords),
        dims=arr.dims,
        coords={arr.dims[i]: coords[i] for i in range(arr.ndim)},
    )
    return out, spl


# def smoothing_spline(arr:xr.DataArray, dim:):

# return scipy.interpolate.make_smoothing_spline()



"""
Base class for implementing loaders that can load data acquired with Scienta Omicron's
DA30L analyzer using ``SES.exe``. Must be subclassed to implement the actual loading.
"""

import configparser
import os
import tempfile
import zipfile

import numpy as np
import xarray as xr

from erlab.io.dataloader import LoaderBase
from erlab.io.igor import load_experiment, load_wave


class DA30Loader(LoaderBase):
    name: str = "_da30"
    aliases: list[str] = []

    name_map: dict[str, str] = {
        "eV": ["Kinetic Energy [eV]", "Energy [eV]"],
        "alpha": ["Y-Scale [deg]", "Thetax [deg]"],
        "beta": ["Thetay [deg]"],
        "hv": ["BL Energy", "Excitation Energy"],
    }
    coordinate_attrs: tuple[str, ...] = ()
    additional_attrs: dict[str, str | int | float] = {}
    always_single: bool = True
    skip_validate: bool = True

    def load_single(self, file_path: str | os.PathLike) -> xr.DataArray:
        ext = os.path.splitext(file_path)[-1]

        match ext:
            case ".ibw":
                data = load_wave(file_path)
            case ".pxt":
                data = load_experiment(file_path)
            case ".zip":
                data = load_zip(file_path)
            case _:
                raise ValueError(f"Unsupported file extension {ext}")

        return self.post_process_general(data)

    def post_process(self, data: xr.DataArray) -> xr.DataArray:
        data = super().post_process(data)

        if "beta" not in data.coords:
            data = data.assign_coords(beta=0.0)

        return data


def load_zip(
    filename: str | os.PathLike,
) -> xr.DataArray | xr.Dataset | list[xr.DataArray]:
    regions: list[str] = []

    with zipfile.ZipFile(filename) as z:
        for fn in z.namelist():
            if fn.startswith("Spectrum_") and fn.endswith(".bin"):
                regions.append(fn[9:-4])

        out = []
        for region in regions:
            with tempfile.TemporaryDirectory() as tmp_dir:
                z.extract(f"Spectrum_{region}.ini", tmp_dir)
                z.extract(f"{region}.ini", tmp_dir)
                z.extract(f"Spectrum_{region}.bin", tmp_dir)

                region_info = parse_ini(
                    os.path.join(tmp_dir, f"Spectrum_{region}.ini")
                )["spectrum"]
                attrs = {}
                for d in parse_ini(os.path.join(tmp_dir, f"{region}.ini")).values():
                    attrs.update(d)

                arr = np.fromfile(
                    os.path.join(tmp_dir, f"Spectrum_{region}.bin"), dtype=np.float32
                )

            shape = []
            coords = {}
            for d in ("depth", "height", "width"):
                n = int(region_info[d])
                offset = float(region_info[f"{d}offset"])
                delta = float(region_info[f"{d}delta"])
                shape.append(n)
                coords[region_info[f"{d}label"]] = np.linspace(
                    offset, offset + (n - 1) * delta, n
                )

            arr = xr.DataArray(
                arr.reshape(shape), coords=coords, name=region_info["name"], attrs=attrs
            )
            out.append(arr)

    if len(out) == 1:
        return out[0]

    try:
        # Try to merge the data without conflicts
        return xr.merge(out)
    except:  # noqa: E722
        # On failure, return a list
        return out


def parse_ini(filename: str | os.PathLike) -> dict:
    parser = configparser.ConfigParser(strict=False)
    out = {}
    with open(filename) as f:
        parser.read_file(f)
        for section in parser.sections():
            out[section] = dict(parser.items(section))
    return out



import numpy as np
from erlab.interactive.imagetool.fastbinning import nanmean_funcs, fast_nanmean


def test_fast_nanmean():
    for nd, funcs in nanmean_funcs.items():
        x = np.random.RandomState(42).randn(*((10,) * nd))
        for axis, func in funcs.items():
            if isinstance(axis, frozenset):
                axis = tuple(axis)
            if not np.allclose(np.nanmean(x, axis), fast_nanmean(x, axis)):
                raise AssertionError(
                    f"fast_nanmean failed for {nd}D array "
                    f"with axis {axis} using {func}."
                )



"""Data loading plugins.

The modules in this package provide data loaders for various beamlines and laboratories.
Each module contains a class that subclasses :class:`erlab.io.dataloader.LoaderBase`,
which can be accessed through :attr:`erlab.io.loaders`.

See :doc:`/generated/erlab.io.dataloader` for more information on how to write a custom
loader.

.. currentmodule:: erlab.io.plugins

Modules
=======

.. autosummary::
   :toctree:

   merlin
   ssrl52
   da30
   kriss

"""

import importlib
import os
import traceback

for fname in os.listdir(os.path.dirname(os.path.abspath(__file__))):
    if (
        not fname.startswith(".")
        and not fname.startswith("__")
        and fname.endswith(".py")
    ):
        try:
            importlib.import_module(__name__ + "." + os.path.splitext(fname)[0])
        except Exception as e:
            print(e)
            traceback.print_exc()



"""Generates simple simulated ARPES data from Graphene bands."""

__all__ = ["generate_data", "generate_data_angles"]

import numpy as np
import scipy.ndimage
import xarray as xr

import erlab.analysis.kspace


def func(kvec, a):
    n1 = np.array([a / np.sqrt(3), 0])
    n2 = np.array([-a / 2 / np.sqrt(3), a / 2])
    n3 = np.array([-a / 2 / np.sqrt(3), -a / 2])
    return (
        np.exp(1j * np.dot(kvec, n1))
        + np.exp(1j * np.dot(kvec, n2))
        + np.exp(1j * np.dot(kvec, n3))
    )


def fermi_dirac(E, T):
    return 1 / (np.exp(E / (8.61733e-5 * T)) + 1)


def band(kvec, t, a):
    return -t * np.abs(func(kvec, a))


def spectral_function(w, bareband, Sreal, Simag):
    return Simag / (np.pi * ((w - bareband - Sreal) ** 2 + Simag**2))


def add_fd_norm(image, eV, temp=30, efermi=0, count=1e3):
    if temp != 0:
        image *= fermi_dirac(eV - efermi, temp)[None, None, :]
    image += 0.5e-6
    image /= image.max()
    image *= count / np.mean(image)


def generate_data(
    shape: tuple[int, int, int] = (250, 250, 300),
    krange: float | tuple[float, float] | dict[str, tuple[float, float]] = 0.89,
    Erange: tuple[float, float] = (-0.45, 0.09),
    temp: float = 20.0,
    a: float = 6.97,
    t: float = 0.43,
    bandshift: float = -0.2,
    Sreal: float = 0.0,
    Simag: float = 0.03,
    kres: float = 0.01,
    Eres: float = 2.0e-3,
    noise: bool = True,
    count: int = 10000,
    ccd_sigma: float = 0.6,
) -> xr.DataArray:
    """
    Generate simulated data for a given shape.

    Parameters
    ----------
    shape
        The shape of the generated data, by default (250, 250, 300)
    krange
        Momentum range in inverse angstroms. Can be a single float, a tuple of floats
        representing the range, or a dictionary with ``kx`` and ``ky`` keys mapping to
        tuples representing the range for each dimension, by default 0.89
    Erange
        Binding energy range in electronvolts, by default (-0.45, 0.09)
    temp
        The temperature in Kelvins for the Fermi-Dirac cutoff. If 0, no cutoff is
        applied, by default 20.0
    a
        Tight binding parameter :math:`a`, by default 6.97
    t
        Tight binding parameter :math:`t`, by default 0.43
    bandshift
        The rigid energy shift in eV, by default -0.2
    Sreal
        The real part of the self energy, by default 0.0
    Simag
        The imaginary part of the self energy, by default 0.03
    kres
        Broadening in momentum in inverse angstroms, by default 0.01
    Eres
        Broadening in energy in electronvolts, by default 2.0e-3
    noise
        Whether to add noise to the generated data, by default `True`
    count
        Determines the signal-to-noise ratio when `noise` is `True`, by default 10000
    ccd_sigma
        The sigma value for CCD noise generation when `noise` is `True`, by default 0.6

    Returns
    -------
    xarray.DataArray
        The generated data with coordinates for kx, ky, and eV.

    """
    if isinstance(krange, dict):
        kx = np.linspace(*krange["kx"], shape[0])
        ky = np.linspace(*krange["ky"], shape[1])
    elif not np.iterable(krange):
        kx = np.linspace(-krange, krange, shape[0])
        ky = np.linspace(-krange, krange, shape[1])
    else:
        kx = np.linspace(*krange, shape[0])
        ky = np.linspace(*krange, shape[1])

    eV = np.linspace(*Erange, shape[2])

    dE = eV[1] - eV[0]
    dk = ((kx[1] - kx[0]) + (ky[1] - ky[0])) / 2

    point_iter = np.array(np.meshgrid(kx, ky)).T.reshape(-1, 2)

    Eij = band(point_iter, t, a).reshape(*shape[:-1], 1)

    Akw_p = spectral_function(eV[None, None, :], Eij + bandshift, Sreal, Simag)
    Akw_m = spectral_function(eV[None, None, :], -Eij + bandshift, Sreal, Simag)

    phase = np.angle(func(point_iter, a)).reshape(shape[0], shape[1])
    Akw_p *= (1 + np.cos(phase))[:, :, None]
    Akw_m *= (1 - np.cos(phase))[:, :, None]
    out = Akw_p + Akw_m

    add_fd_norm(out, eV, efermi=0, temp=temp, count=count)

    if noise:
        rng = np.random.default_rng()
        out = rng.poisson(out).astype(float)

    broadened = scipy.ndimage.gaussian_filter(
        out,
        sigma=[kres / dk, kres / dk, Eres / dE],
        truncate=10.0,
    )
    if noise:
        broadened = scipy.ndimage.gaussian_filter(
            rng.poisson(broadened).astype(float), ccd_sigma, truncate=10.0
        )

    return xr.DataArray(broadened, coords={"kx": kx, "ky": ky, "eV": eV})


def generate_data_angles(
    shape: tuple[int, int, int] = (500, 60, 500),
    angrange: float | tuple[float, float] | dict[str, tuple[float, float]] = 15.0,
    Erange: tuple[float, float] = (-0.45, 0.12),
    hv: float = 50.0,
    configuration: (
        erlab.analysis.kspace.AxesConfiguration | int
    ) = erlab.analysis.kspace.AxesConfiguration.Type1,
    temp: float = 20.0,
    a: float = 6.97,
    t: float = 0.43,
    bandshift: float = -0.2,
    Sreal: float = 0.0,
    Simag: float = 0.03,
    angres: float = 0.1,
    Eres: float = 10.0e-3,
    noise: bool = True,
    count: int = 10000,
    ccd_sigma: float = 0.6,
) -> xr.DataArray:
    """
    Generate simulated data for a given shape.

    Parameters
    ----------
    shape
        The shape of the generated data, by default (250, 250, 300)
    angrange
        Angle range in degrees. Can be a single float, a tuple of floats representing
        the range, or a dictionary with ``alpha`` and ``beta`` keys mapping to tuples
        representing the range for each dimension, by default 15.0
    Erange
        Binding energy range in electronvolts, by default (-0.45, 0.12)
    hv
        The photon energy in eV. Note that the sample work function is assumed to be 4.5
        eV, by default 30.0
    configuration
        The experimental configuration, by default Type1DA
    temp
        The temperature in Kelvins for the Fermi-Dirac cutoff. If 0, no cutoff is
        applied, by default 20.0
    a
        Tight binding parameter :math:`a`, by default 6.97
    t
        Tight binding parameter :math:`t`, by default 0.43
    bandshift
        The rigid energy shift in eV, by default -0.2
    Sreal
        The real part of the self energy, by default 0.0
    Simag
        The imaginary part of the self energy, by default 0.03
    angres
        Broadening in angle in degrees, by default 0.01
    Eres
        Broadening in energy in electronvolts, by default 2.0e-3
    noise
        Whether to add noise to the generated data, by default `True`
    count
        Determines the signal-to-noise ratio when `noise` is `True`, by default 10000
    ccd_sigma
        The sigma value for CCD noise generation when `noise` is `True`, by default 0.6

    Returns
    -------
    xarray.DataArray
        The generated data with coordinates for alpha, beta, and eV.

    """

    if isinstance(angrange, dict):
        alpha = np.linspace(*angrange["alpha"], shape[0])
        beta = np.linspace(*angrange["beta"], shape[1])
    elif not np.iterable(angrange):
        alpha = np.linspace(-angrange, angrange, shape[0])
        beta = np.linspace(-angrange, angrange, shape[1])
    else:
        alpha = np.linspace(*angrange, shape[0])
        beta = np.linspace(*angrange, shape[1])

    if not isinstance(configuration, erlab.analysis.kspace.AxesConfiguration):
        configuration = erlab.analysis.kspace.AxesConfiguration(configuration)

    eV = np.linspace(*Erange, shape[2])

    Ekin = hv - 4.5 + eV[None, None, :]
    forward_func, _ = erlab.analysis.kspace.get_kconv_func(
        Ekin, configuration=configuration, angle_params={}
    )

    dE = eV[1] - eV[0]
    dalpha, dbeta = (alpha[1] - alpha[0]), (beta[1] - beta[0])

    a_mesh, b_mesh = np.meshgrid(alpha, beta, indexing="ij")
    kxv, kyv = forward_func(a_mesh[:, :, None], b_mesh[:, :, None])
    point_iter = np.stack([kxv, kyv], axis=3)

    Eij = band(point_iter, t, a)

    Akw_p = spectral_function(eV, Eij + bandshift, Sreal, Simag)
    Akw_m = spectral_function(eV, -Eij + bandshift, Sreal, Simag)

    phase = np.angle(func(point_iter, a))
    Akw_p *= 1 + np.cos(phase)
    Akw_m *= 1 - np.cos(phase)
    out = Akw_p + Akw_m

    add_fd_norm(out, eV, efermi=0, temp=temp, count=count)

    if noise:
        rng = np.random.default_rng()
        out = rng.poisson(out).astype(float)

    out = scipy.ndimage.gaussian_filter(
        out,
        sigma=[angres / dalpha, angres / dbeta, Eres / dE],
        truncate=10.0,
    )
    if noise:
        out = scipy.ndimage.gaussian_filter(
            rng.poisson(out).astype(float), ccd_sigma, truncate=10.0, axes=(0, 2)
        )

    out = xr.DataArray(out, coords={"alpha": alpha, "beta": beta, "eV": eV})
    out = out.assign_coords(xi=0.0, delta=0.0, hv=hv)

    match configuration:
        case (
            erlab.analysis.kspace.AxesConfiguration.Type1DA
            | erlab.analysis.kspace.AxesConfiguration.Type2DA
        ):
            out = out.assign_coords(chi=0.0)

    return out.assign_attrs(
        configuration=int(configuration),
        temp_sample=temp,
        sample_workfunction=4.5,
    )


if __name__ == "__main__":
    # out = generate_data(
    #     shape=(201, 202, 203),
    #     krange=1.4,
    #     Erange=(-0.45, 0.09),
    #     temp=30,
    #     bandshift=-0.2,
    #     count=1000,
    #     noise=True,
    # )
    out = generate_data_angles()
    import erlab.plotting.erplot as eplt

    eplt.itool([out, out.kspace.convert()])
